##
# File: CifFile.py
# Date: 28-Oct-2018 jdw
#
# Older deprecated API
#
##
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"


import logging
import warnings

from mmcif.core.mmciflib import ParseCifSimple

warnings.filterwarnings("ignore", category=DeprecationWarning)
logger = logging.getLogger(__name__)


class CifFile(object):
    """
    CifFile

    New method prototype --

    CifFile* ParseCifSimple(const std::string& fileName,
                            const bool verbose = false,
                            const unsigned int intCaseSense = 0,
                            const unsigned int maxLineLength = CifFile::STD_CIF_LINE_LENGTH,
                            const std::string& nullValue = CifString::UnknownValue,
                            const std::string& parseLogFileName = std::string());

    """

    def __init__(self, fileName, parseLogFileName=None):
        self.__fileName = fileName

        if parseLogFileName is None:
            self.__cifFile = ParseCifSimple(self.__fileName, False, 0, 255, '?', "")
        else:
            self.__cifFile = ParseCifSimple(self.__fileName, False, 0, 255, '?', parseLogFileName)

    def getCifFile(self):
        return (self.__cifFile)

    @classmethod
    def getFileExt(cls):
        return ('cif')

    def write(self, fileName):
        self.__cifFile.Write(fileName)

    @classmethod
    def read(cls, fileName):
        return cls(fileName)



##
# File:    IoAdapterPy.py
# Author:  J. Westbrook
# Date:    16-Jan-2013
# Version: 0.001 Initial version
#
# Updates:
# 30-Jul-2014 jdw Expose column aligmment and  maximum line length as optional writeFile() parameters.
# 22-Jun-2015 jdw add additional formatting control on write method.
# 15-Aug-2016 rps readFile() updated to accept optional "logtag" parameter for consistency with IoAdapterCore API
# 01-Aug-2017 jdw migrate portions to public repo
# 12-Jan-2018 jdw start to unify api features -
#  6-Aug-2018 jdw set default container properties (locator and load_date)
# 25-Aug-2018 jdw use the input locator rather than uncompressed locator name
##
"""
Python implementation of IoAdapterBase class providing read and write
        methods for PDBx/mmCIF data files -

"""
from __future__ import absolute_import

import io
import logging
import sys

from future.utils import raise_from

from mmcif.io.IoAdapterBase import IoAdapterBase
from mmcif.io.PdbxExceptions import PdbxError, SyntaxError
from mmcif.io.PdbxReader import PdbxReader
from mmcif.io.PdbxWriter import PdbxWriter

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"


logger = logging.getLogger(__name__)


class IoAdapterPy(IoAdapterBase):
    """ Python implementation of IoAdapterBase class providing essential read and write methods for mmCIF data files -

    """

    def __init__(self, *args, **kwargs):
        super(IoAdapterPy, self).__init__(*args, **kwargs)

    def readFile(self, inputFilePath, enforceAscii=False, selectList=None, excludeFlag=False, logFilePath=None, outDirPath=None, cleanUp=False, **kwargs):
        """  Parse the data blocks in the input mmCIF format data file into list of data or definition containers.  The data category content within
            each data block is stored a collection of DataCategory objects within each container.

        Args:
            inputFilePath (string): Input file path
            enforceAscii (bool, optional): Flag to requiring ASCII encoding. See encoding error options.
            selectList (List, optional):  List of data category names to be extracted or excluded from the input file (default: select/extract)
            excludeFlag (bool, optional): Flag to indicate selectList should be treated as an exclusion list
            logFilePath (string, optional): Log file path (if not provided this will be derived from the input file.)
            outDirPath (string, optional): Path for translated/reencoded files and default logfiles.
            cleanUp (bool, optional): Flag to automatically remove logs and temporary files on exit.
            **kwargs: Placeholder for missing keyword arguments.

        Returns:
            List of DataContainers: Contents of input file parsed into a list of DataContainer objects.

        """
        if len(kwargs):
            logger.warn("Unsupported keyword arguments %s" % kwargs.keys())
        filePath = str(inputFilePath)
        oPath = outDirPath if outDirPath else '.'
        containerList = []
        if enforceAscii:
            encoding = 'ascii'
        else:
            encoding = 'utf-8'
        try:
            #
            lPath = logFilePath
            if not lPath:
                lPath = self._getDefaultFileName(filePath, fileType='cif-parser-log', outDirPath=oPath)
            #
            self._setLogFilePath(lPath)
            #
            if not self._fileExists(filePath):
                return []
            filePath = self._uncompress(filePath, oPath)
            #
            if sys.version_info[0] > 2:
                with open(filePath, 'r', encoding=encoding) as ifh:
                    pRd = PdbxReader(ifh)
                    pRd.read(containerList, selectList, excludeFlag=excludeFlag)
            else:
                if enforceAscii:
                    with io.open(filePath, 'r', encoding=encoding) as ifh:
                        pRd = PdbxReader(ifh)
                        pRd.read(containerList, selectList, excludeFlag=excludeFlag)
                else:
                    with open(filePath, 'r') as ifh:
                        pRd = PdbxReader(ifh)
                        pRd.read(containerList, selectList, excludeFlag=excludeFlag)
            if cleanUp:
                self._cleanupFile(lPath, lPath)
                self._cleanupFile(filePath != str(inputFilePath), filePath)
            self._setContainerProperties(containerList, locator=str(inputFilePath), load_date=self._getTimeStamp())
        except (PdbxError, SyntaxError) as ex:
            msg = "File %r with %s" % (filePath, str(ex))
            self._appendToLog([msg])
            self._cleanupFile(lPath and cleanUp, lPath)
            if self._raiseExceptions:
                raise_from(ex, None)
                # raise ex from None
        except Exception as e:
            msg = "File %r with %s" % (filePath, str(e))
            self._appendToLog([msg])
            self._cleanupFile(lPath and cleanUp, lPath)
            if self._raiseExceptions:
                raise e
            else:
                logger.error("Failing read for %s with %s" % (filePath, str(e)))
        return containerList

    def getReadDiags(self):
        """ Return diagnostics from last readFile operation. This will NOT be an exhustive list but
        rather the particular failure that raised a parsing exception.
        """
        return self._readLogRecords()

    def writeFile(self, outputFilePath, containerList, maxLineLength=900, enforceAscii=True,
                  lastInOrder=['pdbx_nonpoly_scheme', 'pdbx_poly_seq_scheme', 'atom_site', 'atom_site_anisotrop'], selectOrder=None,
                  columnAlignFlag=True, useStopTokens=False, formattingStep=None, **kwargs):
        """Write input list of data containers to the specified output file path in mmCIF format.

        Args:
            outputFilePath (string): output file path
            containerList (list DataContainer objects, optional)
            maxLineLength (int, optional): Maximum length of output line (content is wrapped beyond this length)
            enforceAscii (bool, optional): Filter output (not implemented - content must be ascii compatible on input)
            lastInOrder (list of category names, optional): Move data categories in this list to end of each data block
            selectOrder (list of category names, optional): Write only data categories on this list.

            columnAlignFlag (bool, optional): Format the output in aligned columns (default=True) (Native Python Only)
            useStopTokens (bool, optional): Include terminating 'stop_' tokens at the end of mmCIF categories (loop_'s) (Native Python only)
            formattingStep (int, optional): The number row samples within each category used to estimate maximum column width for data alignment (Native Python only)
            **kwargs: Placeholder for unsupported key value pairs

        Returns:
            bool: Completion status


        """
        if len(kwargs):
            logger.warn("Unsupported keyword arguments %s" % kwargs.keys())
        try:
            if enforceAscii:
                encoding = 'ascii'
            else:
                encoding = 'utf-8'
            #
            if sys.version_info[0] > 2:
                with open(outputFilePath, "w", encoding=encoding) as ofh:
                    self.__writeFile(ofh, containerList, maxLineLength=maxLineLength, columnAlignFlag=columnAlignFlag,
                                     lastInOrder=lastInOrder, selectOrder=selectOrder, useStopTokens=useStopTokens,
                                     formattingStep=formattingStep, enforceAscii=enforceAscii, cnvCharRefs=self._useCharRefs)
            else:
                if enforceAscii:
                    with io.open(outputFilePath, 'w', encoding=encoding) as ofh:
                        self.__writeFile(ofh, containerList, maxLineLength=maxLineLength, columnAlignFlag=columnAlignFlag,
                                         lastInOrder=lastInOrder, selectOrder=selectOrder, useStopTokens=useStopTokens,
                                         formattingStep=formattingStep, enforceAscii=enforceAscii, cnvCharRefs=self._useCharRefs)
                else:
                    with open(outputFilePath, "wb") as ofh:
                        self.__writeFile(ofh, containerList, maxLineLength=maxLineLength, columnAlignFlag=columnAlignFlag,
                                         lastInOrder=lastInOrder, selectOrder=selectOrder, useStopTokens=useStopTokens,
                                         formattingStep=formattingStep, enforceAscii=enforceAscii, cnvCharRefs=self._useCharRefs)
            return True
        except Exception as ex:
            if self._raiseExceptions:
                raise_from(ex, None)
            else:
                logger.exception("Failing write for %s with %s" % (outputFilePath, str(ex)))
                logger.error("Failing write for %s with %s" % (outputFilePath, str(ex)))

        return False

    def __writeFile(self, ofh, containerList, maxLineLength=900, columnAlignFlag=True,
                    lastInOrder=None, selectOrder=None, useStopTokens=False,
                    formattingStep=None, enforceAscii=False, cnvCharRefs=False):
        """ Internal method mapping arguments to PDBxWriter API.
        """
        #
        pdbxW = PdbxWriter(ofh)
        pdbxW.setUseStopTokens(flag=useStopTokens)
        pdbxW.setMaxLineLength(numChars=maxLineLength)
        pdbxW.setAlignmentFlag(flag=columnAlignFlag)
        pdbxW.setRowPartition(numParts=formattingStep)
        pdbxW.setConvertCharRefs(flag=cnvCharRefs)
        pdbxW.setSetEnforceAscii(enforceAscii)
        pdbxW.write(containerList, lastInOrder=lastInOrder, selectOrder=selectOrder)



##
#
# File:    DictionaryApi.py
# Author:  jdw
# Date:    11-August-2013
# Version: 0.001
#
# Updates:
#  20-Aug-2013  Jdw manage all data sections -
#   2-Oct-2013  Jdw add alternative methods for extracting full parent and child lists
#   6-Oct-2013  Jdw provide ordering options for history and revision dates
#  10-Dec-2013  jdw return sorted unique list of enumerations with details.
#  14-Feb-2014  jdw fix atribute index lookup.
#  08-Mar-2014  jdw add method getEnumerationClosedFlag(self,category,attribute)
#   9-Sep-2018  jdw add priority to method definition constructor
##
"""
Accessors for PDBx/mmCIF dictionaries -

"""
from __future__ import absolute_import

import logging
import sys

from mmcif.api.DataCategory import DataCategory
from mmcif.api.Method import MethodDefinition, MethodReference
from mmcif.api.PdbxContainers import CifName

from six.moves import zip

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"


logger = logging.getLogger(__name__)


class DictionaryApi(object):

    def __init__(self, containerList, consolidate=True, expandItemLinked=False, verbose=False):
        self.__verbose = verbose
        self.__debug = False
        #
        self.__containerList = containerList
        #
        if (consolidate):
            self.__consolidateDefinitions()

        #
        if (expandItemLinked):
            self.__expandLoopedDefinitions()

        self.__fullIndex = {}

        # ---
        #
        # Map category name to the unique list of attributes
        self.__catNameIndex = {}
        # Map category name to the unique list of item names
        self.__catNameItemIndex = {}
        # Full unique list of item names -
        self.__itemNameList = []
        #
        # Map dictionary objects names to definition containers -
        self.__definitionIndex = {}
        #
        # data section/objects of the dictionary by category name -
        self.__dataIndex = {}
        #
        # Map of types id->(regex,primitive_type)
        self.__typesDict = {}
        #
        self.__enumD = {'ENUMERATION_VALUE': ('item_enumeration', 'value'),
                        'ENUMERATION_DETAIL': ('item_enumeration', 'detail'),
                        'ENUMERATION_TUPLE': ('item_enumeration', None),
                        'ITEM_LINKED_PARENT': ('item_linked', 'parent_name'),
                        'ITEM_LINKED_CHILD': ('item_linked', 'child_name'),
                        'DATA_TYPE_CODE': ('item_type', 'code'),
                        'DATA_TYPE_REGEX': ('item_type_list', 'construct'),
                        'DATA_TYPE_PRIMITIVE': ('item_type_list', 'primitive_code'),
                        'ITEM_NAME': ('item', 'name'),
                        'ITEM_CATEGORY_ID': ('item', 'category_id'),
                        'ITEM_MANDATORY_CODE': ('item', 'mandatory_code'),
                        'ITEM_DESCRIPTION': ('item_description', 'description'),
                        'ITEM_UNITS': ('item_units', 'code'),
                        'ITEM_MANDATORY_CODE': ('item', 'mandatory_code'),
                        'ITEM_DEFAULT_VALUE': ('item_default', 'value'),

                        'ITEM_EXAMPLE_CASE': ('item_examples', 'case'),
                        'ITEM_EXAMPLE_DETAIL': ('item_examples', 'detail'),
                        'ITEM_RANGE_MAXIMUM': ('item_range', 'maximum'),
                        'ITEM_RANGE_MINIMUM': ('item_range', 'minimum'),
                        'CATEGORY_KEY_ITEMS': ('category_key', 'name'),
                        'CATEGORY_EXAMPLE_CASE': ('category_examples', 'case'),
                        'CATEGORY_EXAMPLE_DETAIL': ('category_examples', 'detail'),
                        'CATEGORY_MANDATORY_CODE': ('category', 'mandatory_code'),
                        'CATEGORY_DESCRIPTION': ('category', 'description'),
                        #
                        'DATA_TYPE_CODE_NDB': ('ndb_item_type', 'code'),
                        'ITEM_DESCRIPTION_NDB': ('ndb_item_description', 'description'),
                        'ENUMERATION_VALUE_NDB': ('ndb_item_enumeration', 'value'),
                        'ENUMERATION_DETAIL_NDB': ('ndb_item_enumeration', 'detail'),
                        'ITEM_MANDATORY_CODE_NDB': ('ndb_item', 'mandatory_code'),
                        'ITEM_EXAMPLE_CASE_NDB': ('ndb_item_examples', 'case'),
                        'ITEM_EXAMPLE_DETAIL_NDB': ('ndb_item_examples', 'detail'),
                        'ITEM_RANGE_MAXIMUM_NDB': ('ndb_item_range', 'maximum'),
                        'ITEM_RANGE_MINIMUM_NDB': ('ndb_item_range', 'minimum'),
                        'CATEGORY_EXAMPLE_CASE_NDB': ('ndb_category_examples', 'case'),
                        'CATEGORY_EXAMPLE_DETAIL_NDB': ('ndb_category_examples', 'detail'),
                        'CATEGORY_DESCRIPTION_NDB': ('ndb_category_description', 'description'),
                        #
                        'DATA_TYPE_CODE_PDBX': ('pdbx_item_type', 'code'),
                        'ITEM_DESCRIPTION_PDBX': ('pdbx_item_description', 'description'),
                        'ENUMERATION_VALUE_PDBX': ('pdbx_item_enumeration', 'value'),
                        'ENUMERATION_DETAIL_PDBX': ('pdbx_item_enumeration', 'detail'),
                        'ITEM_MANDATORY_CODE_PDBX': ('pdbx_item', 'mandatory_code'),
                        'ITEM_EXAMPLE_CASE_PDBX': ('pdbx_item_examples', 'case'),
                        'ITEM_EXAMPLE_DETAIL_PDBX': ('pdbx_item_examples', 'detail'),
                        'ITEM_RANGE_MAXIMUM_PDBX': ('pdbx_item_range', 'maximum'),
                        'ITEM_RANGE_MINIMUM_PDBX': ('pdbx_item_range', 'minimum'),
                        'CATEGORY_EXAMPLE_CASE_PDBX': ('pdbx_category_examples', 'case'),
                        'CATEGORY_EXAMPLE_DETAIL_PDBX': ('pdbx_category_examples', 'detail'),
                        'CATEGORY_DESCRIPTION_PDBX': ('pdbx_category_description', 'description'),
                        #
                        'CATEGORY_CONTEXT': ('pdbx_category_context', 'type'),
                        'CATEGORY_GROUP': ('category_group', 'id'),
                        'ITEM_CONTEXT': ('pdbx_item_context', 'type'),
                        'ENUMERATION_CLOSED_FLAG': ('pdbx_item_enumeration_details', 'closed_flag'),
                        #
                        'ITEM_RELATED_FUNCTION_CODE': ('item_related', 'function_code'),
                        'ITEM_RELATED_RELATED_NAME': ('item_related', 'related_name'),

                        'ITEM_ALIAS_ALIAS_NAME': ('item_aliases', 'alias_name'),
                        'ITEM_ALIAS_DICTIONARY': ('item_aliases', 'dictionary'),
                        'ITEM_ALIAS_VERSION': ('item_aliases', 'version'),

                        'ITEM_DEPENDENT_DEPENDENT_NAME': ('item_dependent', 'dependent_name'),
                        'ITEM_SUB_CATEGORY_ID': ('item_sub_category', 'id'),
                        'ITEM_TYPE_CONDITIONS_CODE': ('item_type_conditions', 'code')
                        }
        #
        self.__methodDict = {}
        self.__methodIndex = {}
        #
        self.__makeIndex()
        self.__getMethods()
        #
        self.__fullParentD, self.__fullChildD = self.__makeFullParentChildDictionaries()
        #
        # content sections
        self.__dataBlockDict = {}
        self.__dictionaryDict = {}
        self.__subCategoryDict = {}
        self.__categoryGroupDict = {}
        self.__groupIndex = False
        #
        # Data sections -
        #
        self.__dictionaryHistoryList = []
        self.__itemUnitsDict = {}
        self.__itemUnitsConversionList = []
        self.__itemLinkedGroupDict = {}
        self.__itemLinkedGroupItemDict = {}
        #
        self.__getDataSections()
        #

    #
    #  Methods for data sections --
    #
    def getDictionaryVersion(self):
        try:
            return self.__dictionaryDict['version']
        except Exception:
            return None

    def getDictionaryTitle(self):
        try:
            return self.__dictionaryDict['title']
        except Exception:
            return None

    def getDictionaryUpdate(self, order='reverse'):
        """ Get details from the last history element.
        """
        try:
            if order == 'reverse':
                tD = self.__dictionaryHistoryList[-1]
            else:
                tD = self.__dictionaryHistoryList[0]

            return tD['update']

        except Exception:
            return None

    def getDictionaryRevisionCount(self):
        """ Get details from the last history element.
        """
        try:
            return len(self.__dictionaryHistoryList)
        except Exception:
            return 0

    def getDictionaryHistory(self, order='reverse'):
        """ Returns the revision history as a listr of tuples [(version,update,revisionText),...]
        """
        oL = []
        try:
            if order == 'reverse':
                for tD in reversed(self.__dictionaryHistoryList):
                    oL.append((tD['version'], tD['update'], tD['revision']))
            else:
                for tD in self.__dictionaryHistoryList:
                    oL.append((tD['version'], tD['update'], tD['revision']))
        except Exception:
            pass
        return oL

    def __makeCategoryGroupIndex(self):
        catNameList = self.getCategoryList()
        # add categories in group to self.__categoryGroupDict[<groupName>]['categories']
        for catName in catNameList:
            groupNameList = self.getCategoryGroupList(catName)
            #logger.info("Category %s group list %r\n" % (catName,groupNameList))
            for groupName in groupNameList:
                if groupName not in self.__categoryGroupDict:
                    #  handle undefined category group ?
                    tD = {}
                    tD['description'] = None
                    tD['parent_id'] = None
                    tD['categories'] = []
                    self.__categoryGroupDict[groupName] = tD

                self.__categoryGroupDict[groupName]['categories'].append(catName)
        #
        for groupName in self.__categoryGroupDict.keys():
            #logger.info("Group %s count %r\n" % (groupName, len(self.__categoryGroupDict[groupName]['categories'])))
            if 'categories' in self.__categoryGroupDict[groupName]:
                self.__categoryGroupDict[groupName]['categories'].sort()
        self.__groupIndex = True

    #
    def getCategoryGroupDescription(self, groupName):
        try:
            return self.__categoryGroupDict[groupName]['description']
        except Exception:
            return None

    def getCategoryGroupParent(self, groupName):
        try:
            return self.__categoryGroupDict[groupName]['parent_id']
        except Exception:
            return None

    def getCategoryGroupCategories(self, groupName):
        try:
            if not self.__groupIndex:
                self.__makeCategoryGroupIndex()
            return self.__categoryGroupDict[groupName]['categories']
        except Exception:
            logger.exception("DictionaryApi.getCategoryGroupCategories failed for group %s \n" % groupName)

        return []

    def getCategoryGroups(self):
        try:
            kL = sorted(self.__categoryGroupDict.keys())
            return kL
        except Exception:
            return []

    #
    def getParentCategories(self, categoryName):
        itemNameList = self.getItemNameList(categoryName)
        parentCategories = set()
        for itemName in itemNameList:
            categoryName = CifName.categoryPart(itemName)
            attributeName = CifName.attributePart(itemName)
            parentItemList = self.getFullParentList(categoryName, attributeName)
            for parentItem in parentItemList:
                parentCategoryName = CifName.categoryPart(parentItem)
                parentCategories.add(parentCategoryName)
        return list(parentCategories)

    def getChildCategories(self, categoryName):
        itemNameList = self.getItemNameList(categoryName)
        childCategories = set()
        for itemName in itemNameList:
            categoryName = CifName.categoryPart(itemName)
            attributeName = CifName.attributePart(itemName)
            childItemList = self.getFullChildList(categoryName, attributeName)
            for childItem in childItemList:
                childCategoryName = CifName.categoryPart(childItem)
                childCategories.add(childCategoryName)
        return list(childCategories)

    #
    #def XgetItemNameList(self):
    #    return self.__itemNameList

    #
    def definitionExists(self, definitionName):
        if definitionName in self.__definitionIndex:
            return True
        return False

    def getTypeConditionsCode(self, category, attribute):
        return self.__get('ITEM_TYPE_CONDITIONS_CODE', category, attribute)

    def getItemDependentNameList(self, category, attribute):
        return self.__getList('ITEM_DEPENDENT_DEPENDENT_NAME', category, attribute)

    def getItemSubCategoryIdList(self, category, attribute):
        return self.__getList('ITEM_SUB_CATEGORY_ID', category, attribute)

    def getItemAliasList(self, category, attribute):
        aNL = self.__getListAll('ITEM_ALIAS_ALIAS_NAME', category, attribute)
        aDL = self.__getListAll('ITEM_ALIAS_DICTIONARY', category, attribute)
        aVL = self.__getListAll('ITEM_ALIAS_VERSION', category, attribute)
        aL = []
        for aN, aD, aV in zip(aNL, aDL, aVL):
            aL.append((aN, aD, aV))
        return aL

    def getEnumListWithDetail(self, category, attribute):
        eVL = self.__getListAll('ENUMERATION_VALUE', category, attribute)
        eDL = self.__getListAll('ENUMERATION_DETAIL', category, attribute)
        rL = []
        d = {}
        if len(eVL) == len(eDL):
            for eV, eD in zip(eVL, eDL):
                d[eV] = (eV, eD)
        else:
            for eV in eVL:
                d[eV] = (eV, None)
        #
        for ky in sorted(d.keys()):
            rL.append(d[ky])
        return rL

    def getEnumListAltWithDetail(self, category, attribute):
        eVL = self.__getListAll('ENUMERATION_VALUE_PDBX', category, attribute)
        eDL = self.__getListAll('ENUMERATION_DETAIL_PDBX', category, attribute)

        rL = []
        d = {}
        if len(eVL) == len(eDL):
            for eV, eD in zip(eVL, eDL):
                d[eV] = (eV, eD)
        else:
            for eV in eVL:
                d[eV] = (eV, None)
        #
        for ky in sorted(d.keys()):
            rL.append(d[ky])
        #
        if len(rL) < 1:
            return self.getEnumListWithDetail(category, attribute)
        else:
            return rL

    def getItemRelatedList(self, category, attribute):
        rNL = self.__getListAll('ITEM_RELATED_RELATED_NAME', category, attribute)
        rFL = self.__getListAll('ITEM_RELATED_FUNCTION_CODE', category, attribute)
        rL = []
        for rN, rF in zip(rNL, rFL):
            rL.append((rN, rF))
        return rL

    def getTypeCode(self, category, attribute):
        return self.__get('DATA_TYPE_CODE', category, attribute, followAncestors=True)

    def getTypeCodeAlt(self, category, attribute, fallBack=True):
        v = self.getTypeCodePdbx(category, attribute)
        if v is None:
            v = self.getTypeCodeNdb(category, attribute)
        if fallBack and v is None:
            v = self.getTypeCode(category, attribute)
        return v

    def getTypeCodeNdb(self, category, attribute):
        return self.__get('DATA_TYPE_CODE_NDB', category, attribute, followAncestors=False)

    def getTypeCodePdbx(self, category, attribute):
        return self.__get('DATA_TYPE_CODE_PDBX', category, attribute, followAncestors=False)

    def getDefaultValue(self, category, attribute):
        return self.__get('ITEM_DEFAULT_VALUE', category, attribute)

    def getMandatoryCode(self, category, attribute):
        return self.__get('ITEM_MANDATORY_CODE', category, attribute)

    def getMandatoryCodeAlt(self, category, attribute, fallBack=True):
        v = self.getMandatoryCodePdbx(category, attribute)
        if v is None:
            v = self.getMandatoryCodeNdb(category, attribute)
        if fallBack and v is None:
            v = self.getMandatoryCode(category, attribute)
        return v

    def getMandatoryCodeNdb(self, category, attribute):
        return self.__get('ITEM_MANDATORY_CODE_NDB', category, attribute)

    def getMandatoryCodePdbx(self, category, attribute):
        return self.__get('ITEM_MANDATORY_CODE_PDBX', category, attribute)

    def getTypeRegex(self, category, attribute):
        code = self.getTypeCode(category, attribute)
        if code in self.__typesDict:
            return self.__typesDict[code][1]
        return None

    def getTypeRegexAlt(self, category, attribute, fallBack=True):
        v = self.getTypeRegexPdbx(category, attribute)
        if v is None:
            v = self.getTypeRegexNdb(category, attribute)
        if fallBack and v is None:
            v = self.getTypeRegex(category, attribute)
        return v

    def getTypeRegexNdb(self, category, attribute):
        code = self.getTypeCodeNdb(category, attribute)
        if code in self.__typesDict:
            return self.__typesDict[code][1]
        return None

    def getTypeRegexPdbx(self, category, attribute):
        code = self.getTypeCodePdbx(category, attribute)
        if code in self.__typesDict:
            return self.__typesDict[code][1]
        return None

    def getTypePrimitive(self, category, attribute):
        code = self.getTypeCode(category, attribute)
        if code in self.__typesDict:
            return self.__typesDict[code][0]
        return None

    def getTypeDetail(self, category, attribute):
        code = self.getTypeCode(category, attribute)
        if code in self.__typesDict:
            return self.__typesDict[code][2]
        return None

    def getContextList(self, category, attribute):
        return self.__getList('ITEM_CONTEXT', category, attribute)

    def getCategoryContextList(self, category):
        return self.__getList('CATEGORY_CONTEXT', category, attribute=None)

    def getEnumList(self, category, attribute, sortFlag=True):
        if sortFlag:
            return self.__getList('ENUMERATION_VALUE', category, attribute)
        else:
            return self.__getListAll('ENUMERATION_VALUE', category, attribute)

    def getEnumListAlt(self, category, attribute, fallBack=True, sortFlag=True):
        vL = self.getEnumListPdbx(category, attribute, sortFlag=sortFlag)
        if len(vL) < 1:
            vL = self.getEnumListNdb(category, attribute, sortFlag=sortFlag)
        if fallBack and len(vL) < 1:
            vL = self.getEnumList(category, attribute, sortFlag=sortFlag)
        return vL

    def getEnumListNdb(self, category, attribute, sortFlag=True):
        if sortFlag:
            return self.__getList('ENUMERATION_VALUE_NDB', category, attribute)
        else:
            return self.__getListAll('ENUMERATION_VALUE_NDB', category, attribute)

    def getEnumListPdbx(self, category, attribute, sortFlag=True):
        if sortFlag:
            return self.__getList('ENUMERATION_VALUE_PDBX', category, attribute)
        else:
            return self.__getListAll('ENUMERATION_VALUE_PDBX', category, attribute)

    def isEnumerated(self, category, attribute):
        return (len(self.__getList('ENUMERATION_VALUE', category, attribute)) > 0)

    def isEnumeratedAlt(self, category, attribute, fallBack=True):
        eC = len(self.__getList('ENUMERATION_VALUE_PDBX', category, attribute))
        if (eC == 0):
            eC = len(self.__getList('ENUMERATION_VALUE_NDB', category, attribute))
        if (fallBack and (eC == 0)):
            eC = len(self.__getList('ENUMERATION_VALUE', category, attribute))
        return (eC > 0)

    def getEnumerationClosedFlag(self, category, attribute):
        return self.__get('ENUMERATION_CLOSED_FLAG', category, attribute)

    def getUltimateParent(self, category, attribute):
        """  Return the first ultimate parent item for the input item.
        """
        #        pL=self.__getList('ITEM_LINKED_PARENT',category,attribute)
        pL = self.getFullParentList(category, attribute)
        itemName = CifName.itemName(category, attribute)
        while ((len(pL) > 0) and (pL[0] != itemName)):
            attN = CifName.attributePart(pL[0])
            catN = CifName.categoryPart(pL[0])
            itemName = pL[0]
            pL = self.getFullParentList(catN, attN)
            # pL=self.__getList('ITEM_LINKED_PARENT',catN,attN)
        return itemName

    def getParentList(self, category, attribute, stripSelfParent=False):
        if stripSelfParent:
            itemName = CifName.itemName(category, attribute)
            pL = self.__getList('ITEM_LINKED_PARENT', category, attribute)
            if len(pL) > 0:
                try:
                    pL.remove(itemName)
                except Exception:
                    pass
            return pL
        else:
            return self.__getList('ITEM_LINKED_PARENT', category, attribute)

    def getChildList(self, category, attribute):
        return self.__getList('ITEM_LINKED_CHILD', category, attribute)

    def getFullChildList(self, category, attribute):
        try:
            itemName = CifName.itemName(category, attribute)
            return self.__fullChildD[itemName]
        except Exception:
            return []

    def XgetFullParentList(self, category, attribute):
        try:
            itemName = CifName.itemName(category, attribute)
            return self.__fullParentD[itemName]
        except Exception:
            return []

    def getFullParentList(self, category, attribute, stripSelfParent=False):
        try:
            itemName = CifName.itemName(category, attribute)
            pL = self.__fullParentD[itemName]
            if stripSelfParent:
                if len(pL) > 0:
                    try:
                        pL.remove(itemName)
                    except Exception:
                        pass
                return pL
            else:
                return pL
        except Exception:
            return []

    def getUnits(self, category, attribute):
        return self.__get('ITEM_UNITS', category, attribute)

    def getImplicitList(self):
        iL = []
        for name, dL in self.__definitionIndex.items():
            for d in dL:
                type = d.getType()
                if (type == "definition" and d.isAttribute()):
                    catN = CifName.categoryPart(name)
                    attN = CifName.attributePart(name)
                    if (self.__get('ITEM_MANDATORY_CODE', catN, attN) == 'implicit'):
                        if name not in iL:
                            iL.append(name)
        return iL

    def getDescription(self, category, attribute):
        return self.__get('ITEM_DESCRIPTION', category, attribute)

    def getDescriptionAlt(self, category, attribute, fallBack=True):
        v = self.getDescriptionPdbx(category, attribute)
        if v is None:
            v = self.getDescriptionNdb(category, attribute)
        if fallBack and v is None:
            v = self.getDescription(category, attribute)
        return v

    def getDescriptionNdb(self, category, attribute):
        return self.__get('ITEM_DESCRIPTION_NDB', category, attribute)

    def getDescriptionPdbx(self, category, attribute):
        return self.__get('ITEM_DESCRIPTION_PDBX', category, attribute)

    def getExampleList(self, category, attribute):
        exCL = self.__getListAll('ITEM_EXAMPLE_CASE', category, attribute)
        exDL = self.__getListAll('ITEM_EXAMPLE_DETAIL', category, attribute)
        exL = []
        if len(exCL) == len(exDL):
            for exC, exD in zip(exCL, exDL):
                exL.append((exC, exD))
        else:
            for exC in exCL:
                exL.append((exC, None))
        return exL

    def getExampleListAlt(self, category, attribute, fallBack=True):
        vL = self.getExampleListPdbx(category, attribute)
        if len(vL) < 1:
            vL = self.getExampleListNdb(category, attribute)
        if fallBack and len(vL) < 1:
            vL = self.getExampleList(category, attribute)
        return vL

    def getExampleListNdb(self, category, attribute):
        exCL = self.__getListAll('ITEM_EXAMPLE_CASE_NDB', category, attribute)
        exDL = self.__getListAll('ITEM_EXAMPLE_DETAIL_NDB', category, attribute)
        exL = []
        if len(exCL) == len(exDL):
            for exC, exD in zip(exCL, exDL):
                exL.append((exC, exD))
        else:
            for exC in exCL:
                exL.append((exC, None))
        return exL

    def getExampleListPdbx(self, category, attribute):
        exCL = self.__getListAll('ITEM_EXAMPLE_CASE_PDBX', category, attribute)
        exDL = self.__getListAll('ITEM_EXAMPLE_DETAIL_PDBX', category, attribute)
        exL = []
        if len(exCL) == len(exDL):
            for exC, exD in zip(exCL, exDL):
                exL.append((exC, exD))
        else:
            for exC in exCL:
                exL.append((exC, None))
        return exL

    def getBoundaryList(self, category, attribute):
        minL = self.__getListAll('ITEM_RANGE_MINIMUM', category, attribute)
        maxL = self.__getListAll('ITEM_RANGE_MAXIMUM', category, attribute)
        bL = []
        for vMin, vMax in zip(minL, maxL):
            bL.append((vMin, vMax))
        return bL

    def getBoundaryListAlt(self, category, attribute, fallBack=True):
        vL = self.getBoundaryListPdbx(category, attribute)
        if len(vL) < 1:
            vL = self.getBoundaryListNdb(category, attribute)
        if fallBack and len(vL) < 1:
            vL = self.getBoundaryList(category, attribute)
        return vL

    def getBoundaryListNdb(self, category, attribute):
        minL = self.__getListAll('ITEM_RANGE_MINIMUM_NDB', category, attribute)
        maxL = self.__getListAll('ITEM_RANGE_MAXIMUM_NDB', category, attribute)
        bL = []
        for vMin, vMax in zip(minL, maxL):
            bL.append((vMin, vMax))
        #
        return bL

    def getBoundaryListPdbx(self, category, attribute):
        minL = self.__getListAll('ITEM_RANGE_MINIMUM_PDBX', category, attribute)
        maxL = self.__getListAll('ITEM_RANGE_MAXIMUM_PDBX', category, attribute)
        bL = []
        for vMin, vMax in zip(minL, maxL):
            bL.append((vMin, vMax))
        #
        return bL

    def getCategoryKeyList(self, category):
        return self.__getList('CATEGORY_KEY_ITEMS', category, attribute=None)

    def getCategoryGroupList(self, category):
        return self.__getList('CATEGORY_GROUP', category, attribute=None)

    def getCategoryMandatoryCode(self, category):
        return self.__get('CATEGORY_MANDATORY_CODE', category, attribute=None)

    def getCategoryDescription(self, category):
        return self.__get('CATEGORY_DESCRIPTION', category, attribute=None)

    def getCategoryDescriptionAlt(self, category, fallBack=True):
        v = self.getCategoryDescriptionPdbx(category)
        if v is None:
            v = self.getCategoryDescriptionNdb(category)
        if (fallBack and v is None):
            v = self.getCategoryDescription(category)
        return v

    def getCategoryDescriptionNdb(self, category):
        val = self.__get('CATEGORY_DESCRIPTION_NDB', category, attribute=None)
        return val

    def getCategoryDescriptionPdbx(self, category):
        val = self.__get('CATEGORY_DESCRIPTION_PDBX', category, attribute=None)
        return val

    def getCategoryExampleList(self, category):
        exCL = self.__getListAll('CATEGORY_EXAMPLE_CASE', category, attribute=None)
        exDL = self.__getListAll('CATEGORY_EXAMPLE_DETAIL', category, attribute=None)
        exL = []
        if len(exCL) == len(exDL):
            for exC, exD in zip(exCL, exDL):
                exL.append((exC, exD))
        else:
            for exC in exCL:
                exL.append((exC, None))
        return exL

    def getCategoryExampleListAlt(self, category, fallBack=True):
        vL = self.getCategoryExampleListPdbx(category)
        if len(vL) < 1:
            vL = self.getCategoryExampleListNdb(category)
        if fallBack and len(vL) < 1:
            vL = self.getCategoryExampleList(category)
        return vL

    def getCategoryExampleListNdb(self, category):
        exCL = self.__getListAll('CATEGORY_EXAMPLE_CASE_NDB', category, attribute=None)
        exDL = self.__getListAll('CATEGORY_EXAMPLE_DETAIL_NDB', category, attribute=None)
        exL = []
        if len(exCL) == len(exDL):
            for exC, exD in zip(exCL, exDL):
                exL.append((exC, exD))
        else:
            for exC in exCL:
                exL.append((exC, None))
        return exL

    def getCategoryExampleListPdbx(self, category):
        exCL = self.__getListAll('CATEGORY_EXAMPLE_CASE_PDBX', category, attribute=None)
        exDL = self.__getListAll('CATEGORY_EXAMPLE_DETAIL_PDBX', category, attribute=None)
        exL = []
        if len(exCL) == len(exDL):
            for exC, exD in zip(exCL, exDL):
                exL.append((exC, exD))
        else:
            for exC in exCL:
                exL.append((exC, None))

        return exL

    def getParentDictionary(self):
        """ Create a dictionary of parents relations accross all definnitions

            as d[child]=[parent,parent,...]

            Exclude self parents.
        """
        parentD = {}
        pAtN = self.__enumD['ITEM_LINKED_PARENT'][1]
        cAtN = self.__enumD['ITEM_LINKED_CHILD'][1]

        for dObj in self.__containerList:
            dc = dObj.getObj(self.__enumD['ITEM_LINKED_PARENT'][0])
            if dc is not None:
                idxP = dc.getIndex(pAtN)
                idxC = dc.getIndex(cAtN)
                for row in dc.getRowList():
                    pVal = row[idxP]
                    cVal = row[idxC]
                    if pVal == cVal:
                        continue
                    if cVal not in parentD:
                        parentD[cVal] = []
                    parentD[cVal].append(pVal)
        #
        return parentD

    def __makeFullParentChildDictionaries(self):
        """ Create a dictionaries of full parent/child relations accross all definnitions

            as  fullParentD[child]=[parent,parent,...]
            and fullChildD[parent]=[child,child,...]

            Exclude self parents.
        """
        fullParentD = {}
        fullChildD = {}
        pAtN = self.__enumD['ITEM_LINKED_PARENT'][1]
        cAtN = self.__enumD['ITEM_LINKED_CHILD'][1]

        for dObj in self.__containerList:
            # logger.info("\n\nSearching object  %s\n" % dObj.getName())
            dc = dObj.getObj(self.__enumD['ITEM_LINKED_PARENT'][0])
            if dc is not None:
                idxP = dc.getIndex(pAtN)
                idxC = dc.getIndex(cAtN)
                for row in dc.getRowList():
                    pVal = row[idxP]
                    cVal = row[idxC]
                    # logger.info("%s found parent %s child %s \n" % (dObj.getName(),pVal,cVal))
                    if pVal == cVal:
                        continue
                    if cVal not in fullParentD:
                        fullParentD[cVal] = []
                    fullParentD[cVal].append(pVal)
                    #
                    if pVal not in fullChildD:
                        fullChildD[pVal] = []
                    fullChildD[pVal].append(cVal)

        #
        return fullParentD, fullChildD

    def __get(self, enumCode, category, attribute=None, followAncestors=False):
        """  Return the last occurrence of the input dictionary metadata.  If the value
             for the input category/attribute is null/missing then optionally check for
             an ancestor value.
        """
        v0 = self.__getValue(enumCode, category, attribute)
        if not followAncestors:
            return v0
        else:
            if ((v0 is None) or (len(v0) < 1) or (v0 in ['.', '?'])):
                pItem = self.getUltimateParent(category, attribute)
                if ((pItem is not None) and (len(pItem) > 0) and (pItem != CifName.itemName(category, attribute))):
                    if self.__debug:
                        logger.info("DictionaryApi.__get() Reassigning enum code %s  category %s attribute %s to parent %r\n" %
                                    (enumCode, category, attribute, pItem))

                    return self.__getValue(enumCode, CifName.categoryPart(pItem), CifName.attributePart(pItem))

        return v0

    #
    def __getValue(self, enumCode, category, attribute=None):
        """ Returns the last occurrence of the input dictionary metadata (enumCode) for the input category/attribute
            encountered in the list of objects stored at the indicated definition index.

        """
        eS = None
        if enumCode not in self.__enumD:
            return eS

        if (attribute is not None):
            nm = "_" + category + "." + attribute
        else:
            nm = category

        if nm in self.__definitionIndex:
            dObjL = self.__definitionIndex[nm]
            for dObj in dObjL:
                dc = dObj.getObj(self.__enumD[enumCode][0])
                if dc is not None:
                    atN = self.__enumD[enumCode][1]
                    rL = dc.getRowList()
                    if len(rL) > 0:
                        row = rL[0]
                        if atN is not None:
                            if (dc.hasAttribute(atN)):
                                eS = row[dc.getIndex(atN)]
                        else:
                            tL = []
                            for rv in row:
                                tL.append(rv)
                            eS = tL
        return eS

    def __getList(self, enumCode, category, attribute=None):
        ''' Return the unique list of values '''
        eL = []
        if enumCode not in self.__enumD:
            return eL

        if (attribute is not None):
            nm = "_" + category + "." + attribute
        else:
            nm = category

        if nm in self.__definitionIndex:
            dObjL = self.__definitionIndex[nm]
            for dObj in dObjL:
                dc = dObj.getObj(self.__enumD[enumCode][0])
                if dc is not None:
                    atN = self.__enumD[enumCode][1]
                    for row in dc.getRowList():
                        if atN is not None:
                            if (dc.hasAttribute(atN)):
                                eL.append(row[dc.getIndex(atN)])
                        else:
                            tL = []
                            for rv in row:
                                tL.append(rv)
                            eL.append(tL)
        eS = set(eL)
        eL = list(eS)
        return eL

    def __getListAll(self, enumCode, category, attribute=None):
        ''' Return a list of all values  '''
        eL = []
        if enumCode not in self.__enumD:
            return eL

        if (attribute is not None):
            nm = "_" + category + "." + attribute
        else:
            nm = category

        if nm in self.__definitionIndex:
            dObjL = self.__definitionIndex[nm]
            for dObj in dObjL:
                dc = dObj.getObj(self.__enumD[enumCode][0])
                if dc is not None:
                    atN = self.__enumD[enumCode][1]
                    for row in dc.getRowList():
                        if atN is not None:
                            if (dc.hasAttribute(atN)):
                                eL.append(row[dc.getIndex(atN)])
                        else:
                            tL = []
                            for rv in row:
                                tL.append(rv)
                            eL.append(tL)

        return eL

    def getMethodIndex(self):
        return self.__methodIndex

    def __makeIndex(self):
        """  Create indices of definitions, categories and items.
        """
        iD = {}
        for d in self.__containerList:
            name = d.getName()
            type = d.getType()
            #
            if name not in self.__fullIndex:
                self.__fullIndex[name] = []
            self.__fullIndex[name].append(d)
            #
            if (type == "definition" and d.isCategory()):
                if name not in self.__catNameIndex:
                    self.__catNameIndex[name] = []
                if name not in self.__catNameItemIndex:
                    self.__catNameItemIndex[name] = []
                if name not in self.__definitionIndex:
                    self.__definitionIndex[name] = []
                self.__definitionIndex[name].append(d)

            elif (type == "definition" and d.isAttribute()):
                catN = CifName.categoryPart(name)
                attN = CifName.attributePart(name)
                if catN not in self.__catNameItemIndex:
                    self.__catNameItemIndex[catN] = []
                if name not in self.__catNameItemIndex:
                    self.__catNameItemIndex[catN].append(name)

                if catN not in self.__catNameIndex:
                    self.__catNameIndex[catN] = []
                if attN not in self.__catNameIndex[catN]:
                    self.__catNameIndex[catN].append(attN)
                if name not in self.__definitionIndex:
                    self.__definitionIndex[name] = []
                self.__definitionIndex[name].append(d)
                iD[name] = name
            elif (type == "data"):
                for nm in d.getObjNameList():
                    if nm not in self.__dataIndex:
                        self.__dataIndex[nm] = d.getObj(nm)
            else:
                pass
        #
        self.__itemNameList = sorted(iD.keys())

    def getDefinitionIndex(self):
        return self.__definitionIndex

    def getFullIndex(self):
        return self.__fullIndex

    def getMethod(self, id):
        if id in self.__methodDict:
            return self.__methodDict[id]
        else:
            return None

    def getCategoryList(self):
        return sorted(self.__catNameIndex.keys())

    def getCategoryIndex(self):
        return self.__catNameIndex

    def getAttributeNameList(self, category):
        try:
            return sorted(self.__catNameIndex[category])
        except Exception:
            pass
        return []

    def getItemNameList(self, category):
        try:
            return sorted(self.__catNameItemIndex[category])
        except Exception:
            pass
        return []

    def getSubCategoryDescription(self, subCategoryName):
        if subCategoryName in self.__subCategoryDict:
            return self.__subCategoryDict[subCategoryName]
        else:
            return ''

    def __getMethods(self):
        self.__methodDict = {}
        self.__methodIndex = {}
        for ob in self.__containerList:
            if (ob.getType() == 'data'):
                ml = ob.getObj('method_list')
                if ml is not None:
                    # Use row order as priority
                    for ii, row in enumerate(ml.getRowList(), 1):
                        if ml.hasAttribute('id') and ml.hasAttribute('inline'):
                            mth = MethodDefinition(row[ml.getIndex('id')], row[ml.getIndex('code')], row[ml.getIndex('language')], row[ml.getIndex('inline')], ii)
                            self.__methodDict[row[ml.getIndex('id')]] = mth

                ml = ob.getObj('datablock_methods')
                if ml is not None:
                    for row in ml.getRowList():
                        if ml.hasAttribute('method_id'):
                            mth = MethodReference(row[ml.getIndex('method_id')], 'datablock', ob.getName(), None)
                            if (ob.getName() in self.__methodIndex):
                                self.__methodIndex[ob.getName()].append(mth)
                            else:
                                self.__methodIndex[ob.getName()] = []
                                self.__methodIndex[ob.getName()].append(mth)
            elif (ob.getType() == 'definition'):
                mi = ob.getObj('category_methods')
                if mi is not None:
                    for row in mi.getRowList():
                        if mi.hasAttribute('method_id'):
                            mth = MethodReference(row[mi.getIndex('method_id')], 'category', ob.getName(), None)
                            if (ob.getName() in self.__methodIndex):
                                self.__methodIndex[ob.getName()].append(mth)
                            else:
                                self.__methodIndex[ob.getName()] = []
                                self.__methodIndex[ob.getName()].append(mth)
                mi = ob.getObj('item_methods')
                if mi is not None:
                    for row in mi.getRowList():
                        if mi.hasAttribute('method_id'):
                            mth = MethodReference(row[mi.getIndex('method_id')], 'attribute',
                                                  CifName.categoryPart(ob.getName()),
                                                  CifName.attributePart(ob.getName()))
                            if (ob.getName() in self.__methodIndex):
                                self.__methodIndex[ob.getName()].append(mth)
                            else:
                                self.__methodIndex[ob.getName()] = []
                                self.__methodIndex[ob.getName()].append(mth)
            else:
                pass
        return self.__methodIndex

    def dumpCategoryIndex(self, fh=sys.stdout):
        for k, vL in self.__catNameIndex.items():
            uvL = list(set(vL))
            fh.write("Category: %s has %d attributes\n" % (k, len(uvL)))
            for v in sorted(uvL):
                fh.write("  Attribute: %s\n" % v)

    def dumpMethods(self, fh=sys.stdout):
        for k, vL in self.__methodIndex.items():
            fh.write("Method index key: %s length %d\n" % (k, len(vL)))
            for v in vL:
                v.printIt(fh)
        #
        for k, vL in self.__methodIndex.items():
            fh.write("\n------------------------------------\n")
            fh.write("Method inline text: %s\n" % k)
            for v in vL:
                fh.write("Method text: %s\n" % self.getMethod(v.getId()).getInline())

    def dumpEnumFeatures(self, fh=sys.stdout):
        for k, vL in self.__catNameIndex.items():
            uvL = list(set(vL))
            for v in sorted(uvL):
                itL = self.getEnumList(k, v)
                if len(itL) > 0:
                    fh.write("-----------------------------------------------\n")
                    fh.write("       Category : %s\n" % k)
                    fh.write("       Attribute: %s\n" % v)
                    fh.write("     Description: \n%s\n" % self.getDescription(k, v))
                    fh.write("            Type: %s\n" % self.getTypeCode(k, v))
                    fh.write("  Primitive type: %s\n" % self.getTypePrimitive(k, v))
                    fh.write("      Regex type: %s\n" % self.getTypeRegex(k, v))
                    fh.write("      Enum list length %d\n" % len(itL))
                    for it in itL:
                        fh.write("      Enum: %s\n" % it)

    def dumpFeatures(self, fh=sys.stdout):
        for k, vL in self.__catNameIndex.items():
            uvL = list(set(vL))
            fh.write("-----------------------------------------------\n")
            fh.write("Category: %s has %d attributes\n" % (k, len(uvL)))
            fh.write("     Category description: %s\n" % self.getCategoryDescription(k))
            fh.write(" Alt category description: %s\n" % self.getCategoryDescriptionAlt(k))

            fh.write("         Category context: %s\n" % self.getCategoryContextList(k))

            ctL = self.getCategoryExampleList(k)
            if len(ctL) > 0:
                fh.write("    Category example list length %d\n" % len(ctL))
                for ct1, ct2 in ctL:
                    fh.write("      Example   case: %s\n" % ct1)
                    fh.write("      Example detail: %s\n" % ct2)

            ctL = self.getCategoryExampleListAlt(k)
            if len(ctL) > 0:
                fh.write("    Alt category example list length %d\n" % len(ctL))
                for ct1, ct2 in ctL:
                    fh.write("     Alt example   case: %s\n" % ct1)
                    fh.write("     Alt example detail: %s\n" % ct2)

            for v in sorted(uvL):
                fh.write("  Attribute: %s\n" % v)
                fh.write("     Description: %s\n" % self.getDescription(k, v))
                fh.write(" Alt description: %s\n" % self.getDescriptionAlt(k, v))
                fh.write("            Type: %s\n" % self.getTypeCode(k, v))
                fh.write("        Alt Type: %s\n" % self.getTypeCodeAlt(k, v))
                fh.write("  Primitive type: %s\n" % self.getTypePrimitive(k, v))
                fh.write("      Regex type: %s\n" % self.getTypeRegex(k, v))
                fh.write("         Context: %s\n" % self.getContextList(k, v))
                #
                fh.write(" Type conditions: %s\n" % self.getTypeConditionsCode(k, v))
                fh.write("   Subcategories: %s\n" % self.getItemSubCategoryIdList(k, v))
                #
                itL = self.getEnumList(k, v)
                if len(itL) > 0:
                    fh.write("      Enum list length %d\n" % len(itL))
                    for it in itL:
                        fh.write("      Enum: %s\n" % it)

                itL = self.getParentList(k, v)
                if len(itL) > 0:
                    fh.write("    Parent list length %d\n" % len(itL))
                    for it in itL:
                        fh.write("      Parent: %s\n" % it)
                itL = self.getChildList(k, v)
                if len(itL) > 0:
                    fh.write("    Child list length %d\n" % len(itL))
                    for it in itL:
                        fh.write("      Child: %s\n" % it)

                itL = self.getExampleList(k, v)
                if len(itL) > 0:
                    fh.write("    Example list length %d\n" % len(itL))
                    for it1, it2 in itL:
                        fh.write("      Example   case: %s\n" % it1)
                        fh.write("      Example detail: %s\n" % it2)

                itL = self.getBoundaryList(k, v)
                if len(itL) > 0:
                    fh.write("    Boundary list length %d\n" % len(itL))
                    for (it1, it2) in itL:
                        fh.write("      Boundary condition (min,max):  (%s,%s)\n" % (it1, it2))

                itL = self.getEnumListAlt(k, v)
                if len(itL) > 0:
                    fh.write("      Alt enum list length %d\n" % len(itL))
                    for it in itL:
                        fh.write("      Alt enum: %s\n" % it)

                itL = self.getExampleListAlt(k, v)
                if len(itL) > 0:
                    fh.write("    Alt example list length %d\n" % len(itL))
                    for it1, it2 in itL:
                        fh.write("      Alt example   case: %s\n" % it1)
                        fh.write("      Alt example detail: %s\n" % it2)

                itL = self.getBoundaryListAlt(k, v)
                if len(itL) > 0:
                    fh.write("    Alt boundary list length %d\n" % len(itL))
                    for (it1, it2) in itL:
                        fh.write("      Alt boundary condition (min,max):  (%s,%s)\n" % (it1, it2))

                itL = self.getItemRelatedList(k, v)
                if len(itL) > 0:
                    fh.write("    Related name list length %d\n" % len(itL))
                    for (it1, it2) in itL:
                        fh.write("      Related item name   %s function code %s\n" % (it1, it2))

                itL = self.getItemAliasList(k, v)
                if len(itL) > 0:
                    fh.write("    Alias name list length %d\n" % len(itL))
                    for (it1, it2, it3) in itL:
                        fh.write("      Alias name   %s dictionary %s version %s\n" % (it1, it2, it3))

                itL = self.getItemDependentNameList(k, v)
                if len(itL) > 0:
                    fh.write("    Dependent name list length %d\n" % len(itL))
                    for it1 in itL:
                        fh.write("      Dependent item name   %s\n" % it1)

    def dumpDataSections(self, fh=sys.stdout):
        fh.write("Datablock:  %r\n" % list(self.__dataBlockDict.items()))
        fh.write("Dictionary: %r\n" % list(self.__dictionaryDict.items()))
        fh.write("Dictionary History: %r\n" % self.__dictionaryHistoryList)
        fh.write("Subcategories: %r\n" % list(self.__subCategoryDict.items()))
        fh.write("Category groups:  %r\n" % list(self.__categoryGroupDict.items()))
        fh.write("Item units:  %r\n" % list(self.__itemUnitsDict.items()))
        fh.write("Item units conversions: %r \n" % self.__itemUnitsConversionList)
        fh.write("Item linked groups: %r\n" % list(self.__itemLinkedGroupDict.items()))
        fh.write("Item linked group item list: %r\n" % list(self.__itemLinkedGroupItemDict.items()))

    def dumpItemLinkedGroups(self, fh=sys.stdout):
        for categoryId, lgList in self.__itemLinkedGroupDict.items():
            for lg in lgList:
                if (categoryId, lg[1]) in self.__itemLinkedGroupItemDict:
                    fh.write("  Category  %s   linked group %s:\n" % (categoryId, lg[1]))
                    lgIList = self.__itemLinkedGroupItemDict[(categoryId, lg[1])]
                    for lgI in lgIList:
                        fh.write("    group %s --- child item %s   parent item %s\n" % (lg[1], lgI[0], lgI[1]))

    def XcategoryPart(self, name):
        tname = ""
        if name.startswith("_"):
            tname = name[1:]
        else:
            tname = name

        i = tname.find(".")
        if i == -1:
            return tname
        else:
            return tname[:i]

    def XattributePart(self, name):
        i = name.find(".")
        if i == -1:
            return None
        else:
            return name[i + 1:]

    def __newDataCategory(self, categoryName, attributeNameList):
        """  create a new data category -
        """
        aCat = DataCategory(categoryName)
        for attributeName in attributeNameList:
            aCat.appendAttribute(attributeName)
        return aCat

    def __addItemLinkToDef(self, dObj, parent_name, child_name):
        """  Add the input link relationship to the input definition object.
        """
        if dObj.exists('item_linked'):
            # update in place --
            cObj = dObj.getObj('item_linked')
            iFound = False
            idxP = cObj.getIndex('parent_name')
            idxC = cObj.getIndex('child_name')
            for row in cObj.getRowList():
                if parent_name == row[idxP] and child_name == row[idxC]:
                    iFound = True
                    break
            if not iFound:
                nRows = cObj.getRowCount()
                cObj.setValue(child_name, 'child_name', nRows)
                cObj.setValue(parent_name, 'parent_name', nRows)
                if (self.__debug):
                    logger.info("+DictionaryApi.__addItemLinkToDef() appending item link in category %s\n" % dObj.getName())
            return True
        else:
            # create new category and append to input object
            cObj = self.__newDataCategory('item_linked', ['child_name', 'parent_name'])
            cObj.append([child_name, parent_name])
            dObj.append(cObj)
            if (self.__debug):
                logger.info("+DictionaryApi.__addItemLinkToDef() created new item link in category %s\n" % dObj.getName())
            return True

    def __expandLoopedDefinitions(self):
        """  Handle definitions containing looped item and item_linked categories --
        """
        fullIndex = {}
        for d in self.__containerList:
            name = d.getName()
            if name not in fullIndex:
                fullIndex[name] = []
            fullIndex[name].append(d)

        for name, dObjL in fullIndex.items():
            if len(dObjL) > 0:
                ob = dObjL[0]
                if (ob.getType() == 'definition') and ob.exists('item_linked'):
                    cObj = ob.getObj('item_linked')
                    if cObj.getRowCount() > 0:
                        idxP = cObj.getIndex('parent_name')
                        idxC = cObj.getIndex('child_name')
                        itemName = ob.getName()
                        if (self.__debug):
                            logger.info("\n\n+DictionaryApi.__expandLoopedDefinitions() current target item %s\n" % itemName)
                        cObjNext = self.__newDataCategory('item_linked', ['child_name', 'parent_name'])
                        #
                        # Distribute the data for each row --
                        iChanges = 0
                        for row in cObj.getRowList():
                            #
                            parentItemName = row[idxP]
                            childItemName = row[idxC]
                            if parentItemName == childItemName:
                                continue
                            if childItemName != itemName:
                                iChanges += 1
                                if childItemName in fullIndex:
                                    #
                                    # Add this p/c link to the child definition -
                                    #
                                    self.__addItemLinkToDef(fullIndex[childItemName][0], parentItemName, childItemName)
                                else:
                                    # error missing child definition object.
                                    logger.info("+DictionaryApi.__expandLoopedDefinitions() missing child item %s\n" % (childItemName))
                            else:
                                cObjNext.append([row[idxC], row[idxP]])
                        if cObjNext.getRowCount() > 0:
                            ob.replace(cObjNext)
                        else:
                            ob.remove('item_linked')

    def __expandLoopedDefinitionsSAVE(self):
        """  Handle definitions containing looped item and item_linkded categories --
        """
        fullIndex = {}
        for d in self.__containerList:
            name = d.getName()
            if name not in fullIndex:
                fullIndex[name] = []
            fullIndex[name].append(d)

        for name, dObjL in fullIndex.items():
            if len(dObjL) > 0:
                ob = dObjL[0]
                if (ob.getType() == 'definition') and ob.exists('item_linked'):
                    cObj = ob.getObj('item_linked')
                    if cObj.getRowCount() > 1:
                        idxP = cObj.getIndex('parent_name')
                        idxC = cObj.getIndex('child_name')
                        itemName = ob.getName()

                        cObjNext = self.__newDataCategory('item_linked', ['child_name', 'parent_name'])
                        #
                        # Distribute the data for each row --
                        iChanges = 0
                        for row in cObj.getRowList():
                            #
                            parentItemName = row[idxP]
                            childItemName = row[idxC]
                            if childItemName != itemName:
                                iChanges += 1
                                if childItemName in fullIndex:
                                    #
                                    # Add this p/c link to the child definition -
                                    #
                                    self.__addItemLinkToDef(fullIndex[childItemName][0], parentItemName, childItemName)
                                else:
                                    # error missing child definition object.
                                    logger.info("+DictionaryApi.__expandLoopedDefinitions() missing child item %s\n" % (childItemName))
                            else:
                                cObjNext.append([row[idxC], row[idxP]])
                        if cObjNext.getRowCount() > 0 and iChanges > 0:
                            ob.replace(cObjNext)
                        else:
                            ob.remove('item_linked')

        for name, dObjL in fullIndex.items():
            if len(dObjL) > 0:
                ob = dObjL[0]
                if (ob.getType() == 'definition') and ob.exists('item_linked'):
                    cObj = ob.getObj('item_linked')
                    if cObj.getRowCount() > 0:
                        idxP = cObj.getIndex('parent_name')
                        idxC = cObj.getIndex('child_name')
                        itemName = ob.getName()
                        iChanges = 0
                        for row in cObj.getRowList():
                            #
                            parentItemName = row[idxP]
                            childItemName = row[idxC]
                            if childItemName != itemName:
                                logger.info("+DictionaryApi.__expandLoopedDefinitions() item name %s child item %s\n" % (itemName, childItemName))
                                iChanges += 1
                                if childItemName in fullIndex:
                                    #
                                    # Add this p/c link to the child definition -
                                    #
                                    self.__addItemLinkToDef(fullIndex[childItemName][0], parentItemName, childItemName)
                                else:
                                    # error missing child definition object.
                                    logger.info("+DictionaryApi.__expandLoopedDefinitions() missing definition child item %s\n" % (childItemName))
                        if iChanges > 0:
                            ob.remove('item_linked')

    def __consolidateDefinitions(self):
        """ Consolidate definitions into a single save frame section per definition.
        """
        fullIndex = {}
        for d in self.__containerList:
            name = d.getName()
            if name not in fullIndex:
                fullIndex[name] = []
            fullIndex[name].append(d)

        # preserve the original order of sections -
        #
        nList = []
        for dObj in self.__containerList:
            nm = dObj.getName()
            if nm not in nList:
                nList.append(nm)
        #
        for name, dObjL in fullIndex.items():
            if len(dObjL) > 1:
                for d in dObjL[1:]:
                    xList = d.getObjNameList()
                    for nm in xList:
                        if nm not in dObjL[0].getObjNameList():
                            if (self.__debug):
                                logger.info("+DictionaryApi().__consolidateDefinitions() adding %s to %s\n" % (nm, name))
                            catObj = d.getObj(nm)
                            dObjL[0].append(catObj)

        # create a new list of consolidated objects in original list order
        dList = []
        for nm in nList:
            if nm in fullIndex:
                dl = fullIndex[nm]
                dList.append(dl[0])
            else:
                logger.info("+DictionaryApi().__consolidate() missing object name %s\n" % nm)
        # update lists
        self.__containerList = dList

    def getDataTypeList(self):
        """ Return list of tuples containing ('code','primitive_code','construct','detail' )
        """
        rowList = []
        for code in sorted(self.__typesDict.keys()):
            tup = self.__typesDict[code]
            rowList.append((code, tup[0], tup[1], tup[2]))
        return rowList

    def getSubCategoryList(self):
        """ Return list of tuples containing ('id', 'description')
        """
        rowList = []
        for id in sorted(self.__subCategoryDict.keys()):
            description = self.__subCategoryDict[id]
            rowList.append((id, description))
        return rowList

    def getUnitsList(self):
        """ Return list of tuples containing ('id', 'description')
        """
        rowList = []
        for id in sorted(self.__itemUnitsDict.keys()):
            description = self.__itemUnitsDict[id]
            rowList.append((id, description))
        return rowList

    def getUnitsConversionList(self):
        """  Return list of tuples containing ('from_code','to_code','operator','factor')
        """
        return self.__itemUnitsConversionList

    def __getDataSections(self):
        """
        """
        for ob in self.__containerList:

            if (ob.getType() == 'data'):
                if (self.__debug):
                    logger.info("+DictionaryApi().__getDataSections() adding data container name %s  type  %s  \n" % (ob.getName(), ob.getType()))
                #  add detail to data type tuple
                tl = ob.getObj('item_type_list')
                if tl is not None:
                    for row in tl.getRowList():
                        if tl.hasAttribute('code') and tl.hasAttribute('primitive_code') and tl.hasAttribute('construct') and tl.hasAttribute('detail'):
                            self.__typesDict[row[tl.getIndex('code')]] = (row[tl.getIndex('primitive_code')], row[tl.getIndex('construct')], row[tl.getIndex('detail')])

                tl = ob.getObj('datablock')
                if tl is not None:
                    rL = tl.getRowList()
                    if len(rL) > 0:
                        self.__dataBlockDict = {}
                        if tl.hasAttribute('id') and tl.hasAttribute('description'):
                            row = rL[0]
                            self.__dataBlockDict['id'] = row[tl.getIndex('id')]
                            self.__dataBlockDict['description'] = row[tl.getIndex('description')]

                tl = ob.getObj('dictionary')
                if tl is not None:
                    rL = tl.getRowList()
                    if len(rL) > 0:
                        self.__dictionaryDict = {}
                        row = rL[0]
                        if tl.hasAttribute('datablock_id'):
                            self.__dictionaryDict['datablock_id'] = row[tl.getIndex('datablock_id')]
                        if tl.hasAttribute('title'):
                            self.__dictionaryDict['title'] = row[tl.getIndex('title')]
                        if tl.hasAttribute('version'):
                            self.__dictionaryDict['version'] = row[tl.getIndex('version')]

                tl = ob.getObj('dictionary_history')
                if tl is not None:
                    # history as a list of dictionaries -
                    self.__dictionaryHistoryList = []
                    for row in tl.getRowList():
                        if tl.hasAttribute('version') and tl.hasAttribute('revision') and tl.hasAttribute('update'):
                            tD = {}
                            tD['version'] = row[tl.getIndex('version')]
                            tD['revision'] = row[tl.getIndex('revision')]
                            tD['update'] = row[tl.getIndex('update')]
                            self.__dictionaryHistoryList.append(tD)

                tl = ob.getObj('sub_category')
                if tl is not None:
                    # subcategories as a dictionary by id
                    self.__subCategoryDict = {}
                    for row in tl.getRowList():
                        if tl.hasAttribute('id') and tl.hasAttribute('description'):
                            self.__subCategoryDict[row[tl.getIndex('id')]] = row[tl.getIndex('description')]

                tl = ob.getObj('category_group_list')
                if tl is not None:
                    # category groups as a dictionary by id of tuples
                    self.__categoryGroupDict = {}
                    for row in tl.getRowList():
                        if tl.hasAttribute('id') and tl.hasAttribute('description') and tl.hasAttribute('parent_id'):
                            tD = {}
                            tD['description'] = row[tl.getIndex('description')]
                            tD['parent_id'] = row[tl.getIndex('parent_id')]
                            tD['categories'] = []
                            self.__categoryGroupDict[row[tl.getIndex('id')]] = tD

                tl = ob.getObj('item_units_list')
                if tl is not None:
                    # units as a dictionary by code
                    self.__itemUnitsDict = {}
                    for row in tl.getRowList():
                        if tl.hasAttribute('code') and tl.hasAttribute('detail'):
                            self.__itemUnitsDict[row[tl.getIndex('code')]] = row[tl.getIndex('detail')]

                tl = ob.getObj('item_units_conversion')
                if tl is not None:
                    # units conversion as a simple list now
                    self.__itemUnitsConversionList = []
                    for row in tl.getRowList():
                        if tl.hasAttribute('from_code') and tl.hasAttribute('to_code') and tl.hasAttribute('operator') and tl.hasAttribute('factor'):
                            self.__itemUnitsConversionList.append(
                                (row[
                                    tl.getIndex('from_code')], row[
                                    tl.getIndex('to_code')], row[
                                    tl.getIndex('operator')], row[
                                    tl.getIndex('factor')]))

                tl = ob.getObj('pdbx_item_linked_group')
                if tl is not None:
                    # parent-child collections   [category_id] -> [(1,...),(3,...),(4,...) ]
                    self.__itemLinkedGroupDict = {}
                    for row in tl.getRowList():
                        if tl.hasAttribute('category_id') and tl.hasAttribute('link_group_id') and tl.hasAttribute(
                                'label') and tl.hasAttribute('context') and tl.hasAttribute('condition_id'):
                            category_id = row[tl.getIndex('category_id')]
                            if category_id not in self.__itemLinkedGroupDict:
                                self.__itemLinkedGroupDict[category_id] = []
                            self.__itemLinkedGroupDict[category_id].append((row[tl.getIndex('category_id')], row[tl.getIndex('link_group_id')],
                                                                            row[tl.getIndex('context')], row[tl.getIndex('condition_id')]))

                tl = ob.getObj('pdbx_item_linked_group_list')
                if tl is not None:
                    # parent-child collections   [(category_id,link_group_id)] -> [(child_name,parent_name,parent_category),(,...),(,...) ]
                    self.__itemLinkedGroupItemDict = {}
                    for row in tl.getRowList():
                        if tl.hasAttribute('child_category_id') and tl.hasAttribute('link_group_id') and tl.hasAttribute(
                                'child_name') and tl.hasAttribute('parent_name') and tl.hasAttribute('parent_category_id'):
                            child_category_id = row[tl.getIndex('child_category_id')]
                            link_group_id = row[tl.getIndex('link_group_id')]
                            if (child_category_id, link_group_id) not in self.__itemLinkedGroupItemDict:
                                self.__itemLinkedGroupItemDict[(child_category_id, link_group_id)] = []
                            self.__itemLinkedGroupItemDict[
                                (child_category_id, link_group_id)].append(
                                (row[
                                    tl.getIndex('child_name')], row[
                                    tl.getIndex('parent_name')], row[
                                    tl.getIndex('parent_category_id')]))



##
#
# File:     PdbxContainers.py
# Original: 02-Feb-2009   jdw
#
# Update:
#   23-Mar-2011   jdw Added method to rename attributes in category containers.
#   05-Apr-2011   jdw Change cif writer to select double quoting as preferred
#                     quoting style where possible.
#   16-Jan-2012   jdw Create base class for DataCategory class
#   22-Mar-2012   jdw when append attributes to existing categories update
#                     existing rows with placeholder null values.
#    2-Sep-2012   jdw add option to avoid embedded quoting that might
#                     confuse simple parsers.
#    4-Nov-2012   jdw extend static methods in CifName class
#   14-Nov-2012   jdw refactoring
#   28-Jun-2013   jdw expose remove method
#   01-Aug-2017   jdw migrate portions to public repo
#   14-Jan-2018   jdw add method filterObjectNameList(lastInOrder=None, selectOrder=None)
#    4-Apr-2018   jdw adding internal __eq__ and __hash__ methods
#    6-Aug-2018   jdw add setters/getters for container properties
##
"""

A collection of container classes supporting the PDBx/mmCIF storage model.

A base container class is defined which supports common features of
data and definition containers.   PDBx data files are organized in
sections called data blocks which are mapped to data containers.
PDBx dictionaries contain definition sections and data sections
which are mapped to definition and data containes respectively.

Data in both PDBx data files and dictionaries are organized in
data categories. In the PDBx syntax individual items or data
identified by labels of the form '_categoryName.attributeName'.
The terms category and attribute in PDBx jargon are analogous
table and column in relational data model, or class and attribute
in an object oriented data model.

The DataCategory class provides base storage container for instance
data and definition meta data.

"""
from __future__ import absolute_import

import sys

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"


class CifName(object):
    ''' Class of utilities for CIF-style data names -
    '''

    def __init__(self):
        pass

    @staticmethod
    def categoryPart(name):
        tname = ""
        try:
            if name.startswith("_"):
                tname = name[1:]
            else:
                tname = name

            i = tname.find(".")
            if i == -1:
                return tname
            else:
                return tname[:i]
        except Exception:
            return tname

    @staticmethod
    def attributePart(name):
        try:
            i = name.find(".")
            if i == -1:
                return None
            else:
                return name[i + 1:]
        except Exception:
            return None

    @staticmethod
    def itemName(categoryName, attributeName):
        try:
            return '_' + str(categoryName) + '.' + str(attributeName)
        except Exception:
            return None


class ContainerBase(object):
    ''' Container base class for data and definition objects.
    '''

    def __init__(self, name):
        # The enclosing scope of the data container (e.g. data_/save_)
        self.__name = name
        # List of category names within this container -
        self.__objNameList = []
        # dictionary of DataCategory objects keyed by category name.
        self.__objCatalog = {}
        # dictionary for properties of the container
        self.__propCatalog = {}
        self.__type = None

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.__name == other.__name and self.__objNameList == other.__objNameList and self.__objCatalog == other.__objCatalog and self.__type == other.__type and self.__propCatalog == other.__propCatalog

    def __hash__(self):
        return hash((self.__name, tuple(self.__objNameList), self.__type, tuple(self.__objCatalog.items()), tuple(self.__propCatalog.items())))

    def setProp(self, propName, value):
        try:
            self.__propCatalog[propName] = value
            return True
        except Exception:
            return False

    def getProp(self, propName):
        try:
            return self.__propCatalog[propName]
        except Exception:
            return None

    def getType(self):
        return self.__type

    def setType(self, type):
        self.__type = type

    def getName(self):
        return self.__name

    def setName(self, name):
        self.__name = name

    def exists(self, name):
        if name in self.__objCatalog:
            return True
        else:
            return False

    def getObj(self, name):
        if name in self.__objCatalog:
            return self.__objCatalog[name]
        else:
            return None

    def getObjNameList(self):
        return self.__objNameList

    def append(self, obj):
        """ Add the input object to the current object catalog. An existing object
            of the same name will be overwritten.
        """
        if obj.getName() is not None:
            if obj.getName() not in self.__objCatalog:
                # self.__objNameList is keeping track of object order here --
                self.__objNameList.append(obj.getName())
            self.__objCatalog[obj.getName()] = obj

    def replace(self, obj):
        """ Replace an existing object with the input object
        """
        if ((obj.getName() is not None) and (obj.getName() in self.__objCatalog)):
            self.__objCatalog[obj.getName()] = obj

    def printIt(self, fh=sys.stdout, type="brief"):
        fh.write("+ %s container: %30s contains %4d categories\n" %
                 (self.getType(), self.getName(), len(self.__objNameList)))
        for nm in self.__objNameList:
            fh.write("--------------------------------------------\n")
            fh.write("Data category: %s\n" % nm)
            if type == 'brief':
                self.__objCatalog[nm].printIt(fh)
            else:
                self.__objCatalog[nm].dumpIt(fh)

    def rename(self, curName, newName):
        """ Change the name of an object in place -
        """
        try:
            i = self.__objNameList.index(curName)
            self.__objNameList[i] = newName
            self.__objCatalog[newName] = self.__objCatalog[curName]
            self.__objCatalog[newName].setName(newName)
            return True
        except Exception:
            return False

    def remove(self, curName):
        """ Revmove object by name.  Return True on success or False otherwise.
        """
        try:
            if curName in self.__objCatalog:
                del self.__objCatalog[curName]
                i = self.__objNameList.index(curName)
                del self.__objNameList[i]
                return True
            else:
                return False
        except Exception:
            pass

        return False

    def filterObjectNameList(self, lastInOrder=None, selectOrder=None):
        """ Return an ordered list of categories in the input container subject to
            input -

               lastInOrder: list:  categories to be shifted to the end of the container.
               selectOrder: list:  ordered selection of container categories

            returns:
               filNameList: list:  augmented category list or full list (default)
        """
        filNameList = []
        if lastInOrder:
            objNameList = self.__objNameList
            lastList = []
            for nm in objNameList:
                if nm in lastInOrder:
                    lastList.append(nm)
                    continue
                filNameList.append(nm)
            filNameList.extend(lastList)
        elif selectOrder:
            for nm in selectOrder:
                if self.exists(nm):
                    filNameList.append(nm)
        else:
            filNameList = self.__objNameList

        return filNameList

    def toJSON(self):
        return self.__objCatalog


class DefinitionContainer(ContainerBase):

    def __init__(self, name):
        super(DefinitionContainer, self).__init__(name)
        self.setType('definition')
        self.__globalFlag = False

    def isCategory(self):
        if self.exists('category'):
            return True
        return False

    def isAttribute(self):
        if self.exists('item'):
            return True
        return False

    def getGlobal(self):
        return self.__globalFlag

    def printIt(self, fh=sys.stdout, type="brief"):
        fh.write("Definition container: %30s contains %4d categories\n" %
                 (self.getName(), len(self.getObjNameList())))
        if self.isCategory():
            fh.write("Definition type: category\n")
        elif self.isAttribute():
            fh.write("Definition type: item\n")
        else:
            fh.write("Definition type: undefined\n")

        for nm in self.getObjNameList():
            fh.write("--------------------------------------------\n")
            fh.write("Definition category: %s\n" % nm)
            if type == 'brief':
                self.getObj(nm).printIt(fh)
            else:
                self.getObj(nm).dumpId(fh)


class DataContainer(ContainerBase):
    ''' Container class for DataCategory objects.
    '''

    def __init__(self, name):
        super(DataContainer, self).__init__(name)
        self.setType('data')
        self.__globalFlag = False

    def invokeDataBlockMethod(self, type, method, db):
        self.__currentRow = 1
        exec(method.getInline(), globals(), locals())

    def setGlobal(self):
        self.__globalFlag = True

    def getGlobal(self):
        return self.__globalFlag


class SaveFameContainer(ContainerBase):

    def __init__(self, name):
        super(DefinitionContainer, self).__init__(name)
        self.setType('definition')



##
#  File:  mmcif/api/__init__.py
#
#  Date: 9-July-2018 jdw get this py2/3 stuff out of the main code base.
#
#
import sys

__isPy3 = sys.version_info[0] == 3

if __isPy3:
    __STRING_TYPES__ = str
else:
    __STRING_TYPES__ = basestring
#



##
# File:  PdbxExceptions.py
# Date:  2012-01-09  Jdw  Adapted from PdbxParser
#
##
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"


class PdbxError(Exception):
    """ Class for catch general errors
    """

    def __init__(self, msg):
        super(PdbxError, self).__init__(msg)


class SyntaxError(Exception):
    """ Class for catching syntax errors
    """

    def __init__(self, msg):
        super(SyntaxError, self).__init__(msg)
#



##
# File:  PdbxWriter.py
# Date:  2011-10-09 Jdw  Adapted from PdbxParser.py
#
# Updates:
#    5-Apr-2011 jdw  Using the double quote format preference
#   23-Oct-2012 jdw  update path details and reorganize.
#   17-Dec-2012 jdw  Add quoting preference option
#   25-Jul-2014 jdw  expose methods setAlignmentFlag() and setMaxLineLength()
#   23-Jun-2015 jdw  correct misnamed parameter in formatting method
#   28-Dec-2017 jdw  port to
#   13-Jan-2018 jdw  add selection attributes lastInOrder=None, selectOrder=None
#    8-Mar-2018 jdw  make step calc integer division
###
"""
Utilities for writing mmCIF format data and dictionary containers.

"""

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

import logging
import sys

from mmcif.api.DataCategoryFormatted import DataCategoryFormatted
from mmcif.io.PdbxExceptions import PdbxError

logger = logging.getLogger(__name__)


class PdbxWriter(object):

    """Write mmCIF data files or dictionaries using the input container or container list.
    """

    def __init__(self, ofh=sys.stdout):
        self.__ofh = ofh
        self.__containerList = []
        self.__MAXIMUM_LINE_LENGTH = 2048
        self.__SPACING = 2
        self.__INDENT_DEFINITION = 3
        self.__indentSpace = " " * self.__INDENT_DEFINITION
        self.__doDefinitionIndent = False
        # Maximum number of rows checked for value length and format
        self.__rowPartition = None
        # Defaults to double quoting preference -
        self.__preferDoubleQuotes = True
        self.__useAlignedColumns = True
        self.__useStopTokens = False
        self.__cnvCharRefs = False
        #
        self.__enforceAscii = False
        self.__isPy3 = sys.version_info[0] == 3
        # if self.__isPy3:
        #     self.__string_types = str
        # else:
        #    self.__string_types = basestring

    def setSetEnforceAscii(self, bool):
        self.__enforceAscii = bool

    def setConvertCharRefs(self, flag):
        self.__cnvCharRefs = flag

    def setUseStopTokens(self, flag):
        self.__useStopTokens = flag

    def setMaxLineLength(self, numChars):
        self.__MAXIMUM_LINE_LENGTH = numChars

    def setAlignmentFlag(self, flag=True):
        self.__useAlignedColumns = flag

    def setPreferSingleQuotes(self):
        self.__preferDoubleQuotes = False

    def setPreferDoubleQuotes(self):
        self.__preferDoubleQuotes = True

    def setRowPartition(self, numParts):
        ''' Maximum number of partitions used to format value length for column alignment
        '''
        self.__rowPartition = numParts

    def write(self, containerList, lastInOrder=None, selectOrder=None):
        self.__containerList = containerList
        for container in self.__containerList:
            self.writeContainer(container, lastInOrder=lastInOrder, selectOrder=selectOrder)

    def writeContainer(self, container, lastInOrder=None, selectOrder=None):
        indS = " " * self.__INDENT_DEFINITION
        if container.getType() == 'definition':
            self.__write("save_%s\n" % container.getName())
            self.__doDefinitionIndent = True
            self.__write(indS + "#\n")
        elif container.getType() == 'data':
            if (container.getGlobal()):
                self.__write("global_\n")
                self.__doDefinitionIndent = False
                self.__write("\n")
            else:
                self.__write("data_%s\n" % container.getName())
                self.__doDefinitionIndent = False
                self.__write("#\n")

        nmL = container.filterObjectNameList(lastInOrder=lastInOrder, selectOrder=selectOrder)
        for nm in nmL:
            obj = container.getObj(nm)
            objL = obj.getRowList()

            # Skip empty objects
            if len(objL) == 0:
                continue

            # Item - value formattting
            elif len(objL) == 1:
                self.__writeItemValueFormat(obj)

            # Table formatting -
            elif len(objL) > 1 and len(obj.getAttributeList()) > 0:
                if self.__useAlignedColumns:
                    self.__writeTableFormat(obj)
                else:
                    self.__writeTable(obj)
            else:
                raise PdbxError()

            if self.__doDefinitionIndent:
                self.__write(indS + "#")
            else:
                self.__write("#")

        # Add a trailing saveframe reserved word
        if container.getType() == 'definition':
            self.__write("\nsave_\n")
        self.__write("#\n")

    def __write(self, st):
        try:
            if self.__cnvCharRefs:
                self.__ofh.write(st.encode('ascii', 'xmlcharrefreplace').decode('ascii'))
            elif not self.__isPy3:
                if self.__enforceAscii:
                    self.__ofh.write(st.decode('ascii'))
                else:
                    self.__ofh.write(st)
                    # self.__ofh.write(st.encode('utf-8').decode('utf-8'))
            else:
                self.__ofh.write(st)
        except Exception as e:
            logger.exception("write fails with %s for %r" % (str(e), st))

    def __writeItemValueFormat(self, categoryObj):
        # indS = " " * self.__INDENT_DEFINITION
        myCategory = DataCategoryFormatted(categoryObj, preferDoubleQuotes=self.__preferDoubleQuotes)
        # Compute the maximum item name length within this category -
        attributeNameLengthMax = 0
        for attributeName in myCategory.getAttributeList():
            attributeNameLengthMax = max(attributeNameLengthMax, len(attributeName))
        itemNameLengthMax = self.__SPACING + len(myCategory.getName()) + attributeNameLengthMax + 2
        #
        lineList = []
        # lineList.append(indS+"#\n")
        lineList.append("\n")
        for attributeName, iPos in myCategory.getAttributeListWithOrder():
            if self.__doDefinitionIndent:
                #        - add indent --
                lineList.append(self.__indentSpace)

            itemName = "_%s.%s" % (myCategory.getName(), attributeName)
            lineList.append(itemName.ljust(itemNameLengthMax))

            lineList.append(myCategory.getValueFormatted(attributeName, 0))
            lineList.append("\n")

        self.__write("".join(lineList))

    def __writeTableFormat(self, categoryObj):
        indS = " " * self.__INDENT_DEFINITION
        myCategory = DataCategoryFormatted(categoryObj, preferDoubleQuotes=self.__preferDoubleQuotes)
        # Write the declaration of the loop_
        #
        lineList = []
        lineList.append(indS + '#\n')
        if self.__doDefinitionIndent:
            lineList.append(self.__indentSpace)
        lineList.append("loop_")
        for attributeName in myCategory.getAttributeList():
            lineList.append('\n')
            if self.__doDefinitionIndent:
                lineList.append(self.__indentSpace)
            itemName = "_%s.%s" % (myCategory.getName(), attributeName)
            lineList.append(itemName)
        self.__write("".join(lineList))

        #
        # Write the data in tabular format -
        #
        # print myCategory.getName()
        # print myCategory.getAttributeList()

        #    For speed make the following evaluation on a portion of the table
        if self.__rowPartition is not None:
            numSteps = max(1, myCategory.getRowCount() // self.__rowPartition)
        else:
            numSteps = 1

        formatTypeList, dataTypeList = myCategory.getFormatTypeList(steps=numSteps)
        maxLengthList = myCategory.getAttributeValueMaxLengthList(steps=numSteps)
        spacing = " " * self.__SPACING
        #

        # print formatTypeList
        # print dataTypeList
        # print maxLengthList
        #
        for iRow in range(myCategory.getRowCount()):
            lineList = []
            lineList.append('\n')
            if self.__doDefinitionIndent:
                lineList.append(self.__indentSpace + "  ")

            for iAt in range(myCategory.getAttributeCount()):
                formatType = formatTypeList[iAt]
                maxLength = maxLengthList[iAt]

                if (formatType == 'FT_UNQUOTED_STRING' or formatType == 'FT_NULL_VALUE'):
                    val = myCategory.getValueFormattedByIndex(iAt, iRow)
                    lineList.append(val.ljust(maxLength))

                elif formatType == 'FT_NUMBER':
                    val = myCategory.getValueFormattedByIndex(iAt, iRow)
                    lineList.append(val.rjust(maxLength))

                elif formatType == 'FT_QUOTED_STRING':
                    val = myCategory.getValueFormattedByIndex(iAt, iRow)
                    # don't pad the last item in row condition
                    if (iAt == myCategory.getAttributeCount() - 1):
                        lineList.append(val.ljust(len(val)))
                    else:
                        lineList.append(val.ljust(maxLength + 2))

                elif formatType == "FT_MULTI_LINE_STRING":
                    val = myCategory.getValueFormattedByIndex(iAt, iRow)
                    lineList.append(val)

                lineList.append(spacing)

            self.__write("".join(lineList))
        self.__write("\n")
        if self.__useStopTokens:
            self.__write("stop_\n")

    def __writeTable(self, categoryObj, numSteps=5):
        indS = " " * self.__INDENT_DEFINITION
        myCategory = DataCategoryFormatted(categoryObj, preferDoubleQuotes=self.__preferDoubleQuotes)
        # Write the declaration of the loop_
        #
        lineList = []
        lineList.append(indS + '#\n')
        if self.__doDefinitionIndent:
            lineList.append(self.__indentSpace)
        lineList.append("loop_")
        for attributeName in myCategory.getAttributeList():
            lineList.append('\n')
            if self.__doDefinitionIndent:
                lineList.append(self.__indentSpace)
            itemName = "_%s.%s" % (myCategory.getName(), attributeName)
            lineList.append(itemName)
        self.__write("".join(lineList))

        #
        formatTypeList, dataTypeList = myCategory.getFormatTypeList(steps=numSteps)
        spacing = " " * self.__SPACING
        #
        for iRow in range(myCategory.getRowCount()):
            lineList = []
            lineList.append('\n')
            if self.__doDefinitionIndent:
                lineList.append(self.__indentSpace + "  ")

            for iAt in range(myCategory.getAttributeCount()):
                formatType = formatTypeList[iAt]

                if (formatType == 'FT_UNQUOTED_STRING' or formatType == 'FT_NULL_VALUE'):
                    val = myCategory.getValueFormattedByIndex(iAt, iRow)
                    lineList.append(val)

                elif formatType == 'FT_NUMBER':
                    val = myCategory.getValueFormattedByIndex(iAt, iRow)
                    lineList.append(val)

                elif formatType == 'FT_QUOTED_STRING':
                    val = myCategory.getValueFormattedByIndex(iAt, iRow)
                    lineList.append(val)

                elif formatType == "FT_MULTI_LINE_STRING":
                    val = myCategory.getValueFormattedByIndex(iAt, iRow)
                    lineList.append(val)

                lineList.append(spacing)

            self.__write("".join(lineList))
        self.__write("\n")
        if self.__useStopTokens:
            self.__write("stop_\n")



# File: setup.py
# Date: 17-Dec-2017
#
# Update:  17-Jan-2018 jdw - resolve python virtual env issues with Tox.
#           8-Aug-2018 jdw - add py3.7
#
import glob
import os
import platform
import re
import subprocess
import sys
from distutils.version import LooseVersion

from setuptools import Extension, find_packages, setup
from setuptools.command.build_ext import build_ext


class CMakeExtension(Extension):

    def __init__(self, name, sourcedir='', sources=[]):
        Extension.__init__(self, name, sources=sources)
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):

    def run(self):
        try:
            out = subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError("CMake must be installed to build the following extensions: " +
                               ", ".join(e.name for e in self.extensions))

        if platform.system() == "Windows":
            cmake_version = LooseVersion(re.search(r'version\s*([\d.]+)', out.decode()).group(1))
            if cmake_version < '3.1.0':
                raise RuntimeError("CMake >= 3.1.0 is required on Windows")

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):

        #
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        cmake_args = ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + extdir,
                      '-DPYTHON_EXECUTABLE=' + sys.executable]

        # we need to help cmake find the correct python for this virtual env -
        if hasattr(sys, 'real_prefix'):
            lsp = os.path.join(sys.real_prefix, 'lib', 'libpython') + "*"
            isp = os.path.join(sys.real_prefix, 'include', 'python') + "%s.%s" % (sys.version_info.major, sys.version_info.minor) + "*"
        else:
            lsp = os.path.join(sys.exec_prefix, 'lib', 'libpython') + "*"
            isp = os.path.join(sys.exec_prefix, 'include', 'python') + "%s.%s" % (sys.version_info.major, sys.version_info.minor) + "*"
        #
        lpL = glob.glob(lsp)
        if len(lpL):
            lp = lpL[0]
            cmake_args += ['-DPYTHON_LIBRARY=' + lp]

        ipL = glob.glob(isp)
        if len(ipL):
            ip = ipL[0]
            cmake_args += ['-DPYTHON_INCLUDE_DIR=' + ip]
        #
        cfg = 'Debug' if self.debug else 'Release'
        build_args = ['--config', cfg]

        if platform.system() == "Windows":
            cmake_args += ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}'.format(cfg.upper(), extdir)]
            if sys.maxsize > 2**32:
                cmake_args += ['-A', 'x64']
            build_args += ['--', '/m']
        else:
            cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]
            build_args += ['--', '-j2']

        env = os.environ.copy()
        env['CXXFLAGS'] = '{} -DVERSION_INFO=\\"{}\\"'.format(env.get('CXXFLAGS', ''),
                                                              self.distribution.get_version())
        env['RUN_FROM_DISUTILS'] = "yes"
        #
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        #
        if False:
            print("------------------------------")
            print("Extension source path ", ext.sourcedir)
            print("CMAKE_ARGS ", cmake_args)
            print("self.build_temp ", self.build_temp)
            print("extdir", extdir)
            print("ext.name", ext.name)
            print("sys.executable", sys.executable)
            print("sys.exec_prefix", sys.exec_prefix)
            print("CXXFLAGS ", env['CXXFLAGS'])

        #
        subprocess.check_call(['cmake', ext.sourcedir] + cmake_args, cwd=self.build_temp, env=env)
        subprocess.check_call(['cmake', '--build', '.'] + build_args, cwd=self.build_temp)


packages = []
thisPackage = 'mmcif'
requires = ['future', 'six']


with open('mmcif/__init__.py', 'r') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        fd.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError('Cannot find version information')

setup(
    name=thisPackage,
    version=version,
    description='mmCIF Core Access Library',
    long_description="See:  README.md",
    author='John Westbrook',
    author_email='john.westbrook@rcsb.org',
    url='http://mmcif.wwpdb.org',
    #
    license='Apache 2.0',
    classifiers=(
        'Development Status :: 3 - Alpha',
        # 'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ),
    # entry_points={
    #    'console_scripts': [
    #        'onedep_validate_cli=onedep.cli.validate_cli:run',
    #    ]
    # },
    #
    install_requires=['future', 'six'],
    packages=find_packages(exclude=['mmcif.tests', 'tests.*']),
    package_data={
        # If any package contains *.md or *.rst ...  files, include them:
        '': ['*.md', '*.rst', "*.txt", "*.h", "*.C", ".c", "*.cpp"],
    },
    #

    #
    test_suite="mmcif.tests",
    tests_require=['tox'],
    #
    # Not configured ...
    extras_require={
        'dev': ['check-manifest'],
        'test': ['coverage'],
    },
    # Added for
    command_options={
        'build_sphinx': {
            'project': ('setup.py', thisPackage),
            'version': ('setup.py', version),
            'release': ('setup.py', version)
        }
    },
    ext_modules=[CMakeExtension('mmcif.core.mmciflib')],
    cmdclass=dict(build_ext=CMakeBuild),
    zip_safe=False,
)



##
# File:    IoAdapterBase.py
# Author:  J. Westbrook
# Date:    1-Aug-2017
# Version: 0.001 Initial version
#
# Updates:
#   13-Jan-2018 jdw move _getCategoryNameList() PdbxContainerBase class
#    6-Aug-2018 jdw add _setContainerProperties() to assign default container properties
#   10-Aug-2018 jdw add _uncompress() method (suppress xz as lzma is not support in py27)
##
"""
Base class presenting essential PDBx/mmCIF IO methods.

"""

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

import bz2
import datetime
import gzip
import io
import logging
# import lzma
import os
import shutil
import tempfile
import time
import zipfile

from mmcif.io.PdbxExceptions import PdbxError

logger = logging.getLogger(__name__)


class IoAdapterBase(object):
    """ Base class presenting essential mmCIF I/O methods.
    """

    def __init__(self, *args, **kwargs):
        """General options controlling I/O method operations:

        Args:
            raiseExceptions (bool, optional): Flag to indicate that API errors should generate exceptions (True) or catch and log errors (default=False)
            maxInputLineLength (int, optional):  Default maximum input line length (default=4096)
            useCharRefs (bool, optional): Replace non-ascii characters with XML Character References (default=True)
            timing (bool, optional):  log timing details for parsing and processing steps (default=False)
            verbose (bool,optional):  log verbose output from wrapped libraries

        """
        self._raiseExceptions = kwargs.get('raiseExceptions', False)
        self._maxInputLineLength = kwargs.get('maxInputLineLength', 4096)
        self._useCharRefs = kwargs.get('useCharRefs', True)
        self.__logFilePath = None
        self._debug = kwargs.get('debug', False)
        self._timing = kwargs.get('timing', False)
        self._verbose = kwargs.get('verbose', True)

    def readFile(self, inputFilePath, **kwargs):
        """ Read file method. (abstract)

         Args:
            inputFilePath (string):  Input file path/uri
                   kw:  optional key-value arguments

        Returns:
            list of DataContainer Objects:  list of data or definition container objects
        """
        raise NotImplementedError("To be implemented in subclass")

    def writeFile(self, outputFilePath, containerList, **kwargs):
        """ Write file method - (abstract)

        Args:
            outputFilePath (string):  output file path
            containerList (list of DataContainer objects):  list of data or definition containers objects for output

        Returns:
            bool: Completion status
        """
        raise NotImplementedError("To be implemented in subclass")

    def getReadDiags(self):
        """ Return any diagnostics from the last read operation. (abstract)
        """
        raise NotImplementedError("To be implemented in subclass")

    def _setContainerProperties(self, containerList, **kwargs):
        try:
            for container in containerList:
                for kw in kwargs:
                    container.setProp(kw, kwargs[kw])
        except Exception:
            return False

    def _getTimeStamp(self):
        utcnow = datetime.datetime.utcnow()
        ts = utcnow.strftime("%Y-%m-%d:%H:%M:%S")
        return ts

    def _getCategoryNameList(self, container, lastInOrder=None, selectOrder=None):
        """ Return an ordered list of categories in the input container subject to
            input category name lists.

            Args:
               container (DataContainer object):  Input DataContainer object
               lastInOrder (list):  names of categories to be shifted to the end of the container.
               selectOrder (list):  preferred order of category names

            Returns:
               catNameList: list:  augmented category list or full list (default)
        """
        catNameList = []
        if lastInOrder:
            objNameList = container.getObjNameList()
            lastList = []
            for nm in objNameList:
                if nm in lastInOrder:
                    lastList.append(nm)
                    continue
                catNameList.append(nm)
            catNameList.extend(lastList)
        elif selectOrder:
            for nm in selectOrder:
                if container.exists(nm):
                    catNameList.append(nm)
        else:
            catNameList = objNameList

        return catNameList

    def _setLogFilePath(self, filePath):
        """ Set the log file path.
        """
        self.__logFilePath = filePath

    def _getLogFilePath(self):
        """ Return current log file path.
        """
        return self.__logFilePath

    def _appendToLog(self, stList):
        """ Append input string list to the current log file -
        """
        if not self.__logFilePath:
            return
        try:
            with open(self.__logFilePath, 'a') as ofh:
                ofh.write("%s\n" % '\n'.join(stList))
        except Exception as e:
            pass

    def _logError(self, msg):
        """ Convenience method to log error messages and optionally raise general exceptions (PdbxError).
        """
        self._appendToLog([msg])
        if self._raiseExceptions:
            raise PdbxError(msg)
        else:
            logger.error(msg)

    def _readLogRecords(self):
        """ Return the contents of the current log file as list of strings.
        """
        diagL = []
        try:
            with open(self.__logFilePath, 'r') as ifh:
                for line in ifh:
                    diagL.append(line[:-1])
        except Exception as e:
            msg = "No logfile found %s" % self.__logFilePath
            diagL.append(msg)
            logger.debug(msg)

        return diagL

    def __getDiscriminator(self):
        """ Internal method returning a string which can discriminate among default file names -
        """
        return str(int(time.time() * 10000))

    def _getDefaultFileName(self, filePath, fileType='cif-parser', fileExt='log', outDirPath=None, verify=True):
        """ Return default file path for the target input file subject to input attributes and the output path.
        """
        returnFilePath = None
        try:
            dn, fn = os.path.split(filePath)
            bn, ext = os.path.splitext(fn)
            #
            ft = fileType if fileType else 'temp'
            fex = fileExt if fileExt else 'tmp'
            #
            sf = '_' + ft + '_P' + self.__getDiscriminator() + '.' + fex
            #
            pth = outDirPath if outDirPath else '.'
            #
            if verify:
                # test if pth is actually writable ?  Throw exception otherwise -
                #
                testfile = tempfile.TemporaryFile(dir=pth)
                testfile.close()
                #
            returnFilePath = os.path.join(pth, bn + sf)
        except Exception as e:
            if self._raiseExceptions:
                raise e
            else:
                logger.error("Failed creating default filename for %s type %s with %s" % (filePath, fileType, str(e)))

        return returnFilePath

    def _fileExists(self, filePath):
        """ Verify that input file path exists and is readable.
        """
        try:
            if (not os.access(filePath, os.R_OK)):
                msg = "Missing file %r" % filePath
                self._appendToLog([msg])
                logger.error(msg)
                #
                if self._raiseExceptions:
                    raise PdbxError(msg)
                return False
            else:
                logger.debug("Reading from file path %s" % filePath)
                return True
        except Exception as e:
            msg = "File check error for %r with %s " % (filePath, str(e))
            self._appendToLog([msg])
            if self._raiseExceptions:
                raise PdbxError(msg)
            else:
                logger.error(msg)
        return False

    def _cleanupFile(self, test, filePath):
        """  Remove the input file path subject to the input test condition.
        """
        try:
            if test:
                os.remove(filePath)
        except Exception:
            pass

    def _toAscii(self, inputFilePath, outputFilePath, chunkSize=5000, encodingErrors='ignore'):
        """ Encode input file to Ascii and write this to the target output file.   Handle encoding
            errors according to the input settting ('ignore', 'escape', 'xmlcharrefreplace').
        """
        try:
            startTime = time.time()
            chunk = []
            with io.open(inputFilePath, "r", encoding="utf-8") as r, io.open(outputFilePath, "w", encoding='ascii') as w:
                for line in r:
                    # chunk.append(line.encode('ascii', 'xmlcharrefreplace').decode('ascii'))
                    chunk.append(line.encode('ascii', encodingErrors).decode('ascii'))
                    if len(chunk) == chunkSize:
                        w.writelines(chunk)
                        chunk = []
                w.writelines(chunk)
            if self._timing:
                stepTime1 = time.time()
                logger.info("Timing text file %s encoded to as ascii in %.4f seconds" % (inputFilePath, stepTime1 - startTime))
            return True
        except Exception as e:
            msg = "Failing text ascii encoding for %s with %s" % (inputFilePath, str(e))
            self._appendToLog([msg])
            logger.error(msg)
            if self._raiseExceptions:
                raise PdbxError(msg)
        #
        return False

    def _uncompress(self, inputFilePath, outputDir):
        """ Uncompress the input file if the path name has a recognized compression type file extension.file

            Return the file name ofthe uncompressed file (in outDir) or the original input file path.

        """
        try:
            startTime = time.time()
            fp, fn = os.path.split(inputFilePath)
            bn, fx = os.path.splitext(fn)
            outputFilePath = os.path.join(outputDir, bn)
            if inputFilePath.endswith(".gz"):
                with gzip.open(inputFilePath, mode='rb') as inp_f:
                    with io.open(outputFilePath, "wb") as out_f:
                        shutil.copyfileobj(inp_f, out_f)
            elif inputFilePath.endswith(".bz2"):
                with bz2.open(inputFilePath, mode='rb') as inp_f:
                    with io.open(outputFilePath, "wb") as out_f:
                        shutil.copyfileobj(inp_f, out_f)
            #elif inputFilePath.endswith(".xz"):
            #    with lzma.open(inputFilePath, mode="rb") as inp_f:
            #        with io.open(outputFilePath, "wb") as out_f:
            #            shutil.copyfileobj(inp_f, out_f)
            elif inputFilePath.endswith(".zip"):
                with zipfile.ZipFile(inputFilePath, mode='rb') as inp_f:
                    with io.open(outputFilePath, "wb") as out_f:
                        shutil.copyfileobj(inp_f, out_f)
            else:
                outputFilePath = inputFilePath
            if self._timing:
                stepTime1 = time.time()
                logger.info("Timing text file %s uncompressed in %.4f seconds" % (inputFilePath, stepTime1 - startTime))
            #
        except Exception as e:
            msg = "Failing uncompress for file %s with %s" % (inputFilePath, str(e))
            self._appendToLog([msg])
            logger.exception(msg)
            if self._raiseExceptions:
                raise PdbxError(msg)

        logger.debug("Returning file path %r" % outputFilePath)
        return outputFilePath



##
#
# File:     DataCategoryFormatted.py
# Original: 02-Feb-2009   jdw
#
# Update:
#   14-Nov-2012   jdw refactoring
#   17-Nov-2012   jdw self._rowList becomes data -
#   17-Dec-2012   jdw add quoting preference as constructor option.
#   01-Aug-2017   jdw migrate portions to public repo
##
"""

A subclass of DataCategory including additional formatting methods.

"""
from __future__ import absolute_import

import logging
import re

from mmcif.api.DataCategory import DataCategory

from six.moves import range

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"


logger = logging.getLogger(__name__)


class DataCategoryFormatted(DataCategory):
    """  A subclass of DataCategory including additional formatting methods.
    """

    def __init__(self, dataCategoryObj, preferDoubleQuotes=True):
        self.__dcObj = dataCategoryObj
        super(DataCategory, self).__init__(self.__dcObj._name, self.__dcObj._attributeNameList, self.__dcObj.data)
        #
        self._currentRowIndex = 0
        self._currentAttribute = None
        #
        self.__avoidEmbeddedQuoting = False
        self.__preferDoubleQuotes = preferDoubleQuotes
        #
        # --------------------------------------------------------------------
        # any whitespace
        self.__wsRe = re.compile(r"\s")
        # self.__wsAndQuotesRe=re.compile(r"[\s'\"]")
        self.__wsAndQuotesRe = re.compile(r"[\s'\"#]")
        # any newline or carriage control
        self.__nlRe = re.compile(r"[\n\r]")
        #
        # single quote
        self.__sqRe = re.compile(r"[']")
        #
        self.__sqWsRe = re.compile(r"('\s)|(\s')")

        # double quote
        self.__dqRe = re.compile(r'["]')
        self.__dqWsRe = re.compile(r'("\s)|(\s")')
        #
        self.__intRe = re.compile(r'^[0-9]+$')
        self.__floatRe = re.compile(r'^-?(([0-9]+)[.]?|([0-9]*[.][0-9]+))([(][0-9]+[)])?([eE][+-]?[0-9]+)?$')
        #
        self.__dataTypeList = ['DT_NULL_VALUE', 'DT_INTEGER', 'DT_FLOAT', 'DT_UNQUOTED_STRING', 'DT_ITEM_NAME',
                               'DT_DOUBLE_QUOTED_STRING', 'DT_SINGLE_QUOTED_STRING', 'DT_MULTI_LINE_STRING']
        self.__formatTypeList = ['FT_NULL_VALUE', 'FT_NUMBER', 'FT_NUMBER', 'FT_UNQUOTED_STRING',
                                 'FT_QUOTED_STRING', 'FT_QUOTED_STRING', 'FT_QUOTED_STRING', 'FT_MULTI_LINE_STRING']
        #
        # try:
        #    basestring
        # except NameError:
        #    basestring = str
        #
        # self.__string_types = basestring

    def __formatPdbx(self, inp):
        """ Format input data following PDBx quoting rules -
        """
        try:
            if (inp is None):
                return ("?", 'DT_NULL_VALUE')

            # pure numerical values are returned as unquoted strings
            # if (isinstance(inp, int) or self.__intRe.search(str(inp))):
            #
            try:
                if (isinstance(inp, int) or self.__intRe.search(inp)):
                    return ([str(inp)], 'DT_INTEGER')
            except Exception:
                pass

            # if (isinstance(inp, float) or self.__floatRe.search(str(inp))):
            try:
                if (isinstance(inp, float) or self.__floatRe.search(inp)):
                    return ([str(inp)], 'DT_FLOAT')
            except Exception:
                pass

            # null value handling -

            if (inp == "." or inp == "?"):
                return ([inp], 'DT_NULL_VALUE')

            if (inp == ""):
                return (["."], 'DT_NULL_VALUE')

            # Contains white space or quotes ?
            if not self.__wsAndQuotesRe.search(inp):
                # if inp.startswith("_"):
                if inp[0] in ['_']:
                    return (self.__doubleQuotedList(inp), 'DT_ITEM_NAME')
                elif inp[0] in ['[', ']', '$', '#', ';']:
                    return (self.__doubleQuotedList(inp), 'DT_DOUBLE_QUOTED_STRING')
                elif inp[:5] in ['data_', 'loop_', 'save_']:
                    return (self.__doubleQuotedList(inp), 'DT_DOUBLE_QUOTED_STRING')
                else:
                    return ([str(inp)], 'DT_UNQUOTED_STRING')
            else:
                if self.__nlRe.search(inp):
                    return (self.__semiColonQuotedList(inp), 'DT_MULTI_LINE_STRING')
                else:
                    if (self.__preferDoubleQuotes):
                        if (self.__avoidEmbeddedQuoting):
                            # change priority to choose double quoting where possible.
                            if not self.__dqRe.search(inp) and not self.__sqWsRe.search(inp):
                                return (self.__doubleQuotedList(inp), 'DT_DOUBLE_QUOTED_STRING')
                            elif not self.__sqRe.search(inp) and not self.__dqWsRe.search(inp):
                                return (self.__singleQuotedList(inp), 'DT_SINGLE_QUOTED_STRING')
                            else:
                                return (self.__semiColonQuotedList(inp), 'DT_MULTI_LINE_STRING')
                        else:
                            # change priority to choose double quoting where possible.
                            if not self.__dqRe.search(inp):
                                return (self.__doubleQuotedList(inp), 'DT_DOUBLE_QUOTED_STRING')
                            elif not self.__sqRe.search(inp):
                                return (self.__singleQuotedList(inp), 'DT_SINGLE_QUOTED_STRING')
                            else:
                                return (self.__semiColonQuotedList(inp), 'DT_MULTI_LINE_STRING')
                    else:
                        if (self.__avoidEmbeddedQuoting):
                            # change priority to choose double quoting where possible.
                            if not self.__sqRe.search(inp) and not self.__dqWsRe.search(inp):
                                return (self.__singleQuotedList(inp), 'DT_SINGLE_QUOTED_STRING')
                            elif not self.__dqRe.search(inp) and not self.__sqWsRe.search(inp):
                                return (self.__doubleQuotedList(inp), 'DT_DOUBLE_QUOTED_STRING')
                            else:
                                return (self.__semiColonQuotedList(inp), 'DT_MULTI_LINE_STRING')
                        else:
                            # change priority to choose double quoting where possible.
                            if not self.__sqRe.search(inp):
                                return (self.__singleQuotedList(inp), 'DT_SINGLE_QUOTED_STRING')
                            elif not self.__dqRe.search(inp):
                                return (self.__doubleQuotedList(inp), 'DT_DOUBLE_QUOTED_STRING')
                            else:
                                return (self.__semiColonQuotedList(inp), 'DT_MULTI_LINE_STRING')
        except Exception as e:
            logger.exception("Failing with %s on input %r %r" % (str(e), inp, type(inp)))

        return ("?", 'DT_NULL_VALUE')

    def __dataTypePdbx(self, inp):
        """ Detect the PDBx data type -
        """
        if (inp is None):
            return ('DT_NULL_VALUE')

        # pure numerical values are returned as unquoted strings
        # if isinstance(inp, int) or self.__intRe.search(str(inp)):
        if isinstance(inp, int) or self.__intRe.search(inp):
            return ('DT_INTEGER')

        # if isinstance(inp, float) or self.__floatRe.search(str(inp)):
        if isinstance(inp, float) or self.__floatRe.search(inp):
            return ('DT_FLOAT')

        # null value handling -

        if (inp == "." or inp == "?"):
            return ('DT_NULL_VALUE')

        if (inp == ""):
            return ('DT_NULL_VALUE')

        # Contains white space or quotes ?
        if not self.__wsAndQuotesRe.search(inp):
            if inp.startswith("_"):
                return ('DT_ITEM_NAME')
            else:
                return ('DT_UNQUOTED_STRING')
        else:
            if self.__nlRe.search(inp):
                return ('DT_MULTI_LINE_STRING')
            else:
                if (self.__avoidEmbeddedQuoting):
                    if not self.__sqRe.search(inp) and not self.__dqWsRe.search(inp):
                        return ('DT_DOUBLE_QUOTED_STRING')
                    elif not self.__dqRe.search(inp) and not self.__sqWsRe.search(inp):
                        return ('DT_SINGLE_QUOTED_STRING')
                    else:
                        return ('DT_MULTI_LINE_STRING')
                else:
                    if not self.__sqRe.search(inp):
                        return ('DT_DOUBLE_QUOTED_STRING')
                    elif not self.__dqRe.search(inp):
                        return ('DT_SINGLE_QUOTED_STRING')
                    else:
                        return ('DT_MULTI_LINE_STRING')

    def __singleQuotedList(self, inp):
        ll = []
        ll.append("'")
        ll.append(inp)
        ll.append("'")
        return(ll)

    def __doubleQuotedList(self, inp):
        ll = []
        ll.append('"')
        ll.append(inp)
        ll.append('"')
        return(ll)

    def __semiColonQuotedList(self, inp):
        ll = []
        ll.append("\n")
        if inp[-1] == '\n':
            ll.append(";")
            ll.append(inp)
            ll.append(";")
            ll.append("\n")
        else:
            ll.append(";")
            ll.append(inp)
            ll.append("\n")
            ll.append(";")
            ll.append("\n")

        return(ll)

    def getValueFormatted(self, attributeName=None, rowIndex=None):
        if attributeName is None:
            attribute = self._currentAttribute
        else:
            attribute = attributeName

        if rowIndex is None:
            rowI = self._currentRowIndex
        else:
            rowI = rowIndex

        if isinstance(attribute, self._string_types) and isinstance(rowI, int):
            try:
                list, type = self.__formatPdbx(self.data[rowI][self._attributeNameList.index(attribute)])
                return "".join(list)
            except (IndexError):
                logger.exception("attributeName %s rowI %r rowdata %r\n" % (attributeName, rowI, self.data[rowI]))
                raise IndexError
            except Exception as e:
                logger.exception(" Failing with %s - AttributeName %s rowI %r rowdata %r\n" % (str(e), attributeName, rowI, self.data[rowI]))
        else:
            logger.error(" Type error - AttributeName %r rowI %r rowdata %r" % (attributeName, rowI, self.data[rowI]))
            logger.error(" Type error - string types %r" % (self._string_types))
            raise TypeError(attribute)

    def getValueFormattedByIndex(self, attributeIndex, rowIndex):
        try:
            list, type = self.__formatPdbx(self.data[rowIndex][attributeIndex])
            return "".join(list)
        except (IndexError):
            logger.exception("attributeIndex %r rowIndex %r rowdata %r\n" % (attributeIndex, rowIndex, self.data[rowIndex][attributeIndex]))
            raise IndexError
        except Exception as e:
            logger.exception("Failing with %s  - attributeIndex %r rowIndex %r rowdata %r\n" % (str(e), attributeIndex, rowIndex, self.data[rowIndex][attributeIndex]))
            raise e

    def getAttributeValueMaxLengthList(self, steps=1):
        mList = [0 for i in range(len(self._attributeNameList))]
        for row in self.data[::steps]:
            for indx in range(len(self._attributeNameList)):
                val = row[indx]
                if isinstance(val, self._string_types):
                    tLen = len(val)
                else:
                    tLen = len(str(val))

                mList[indx] = max(mList[indx], tLen)
        return mList

    def getFormatTypeList(self, steps=1):
        try:
            curFormatTypeList = []
            curDataTypeList = ['DT_NULL_VALUE' for i in range(len(self._attributeNameList))]
            for row in self.data[::steps]:
                for indx in range(len(self._attributeNameList)):
                    val = row[indx]
                    # print "index ",indx," val ",val
                    dType = self.__dataTypePdbx(val)
                    dIndx = self.__dataTypeList.index(dType)
                    # print "d type", dType, " d type index ",dIndx

                    cType = curDataTypeList[indx]
                    cIndx = self.__dataTypeList.index(cType)
                    cIndx = max(cIndx, dIndx)
                    curDataTypeList[indx] = self.__dataTypeList[cIndx]

            # Map the format types to the data types
            curFormatTypeList = []
            for dt in curDataTypeList:
                ii = self.__dataTypeList.index(dt)
                curFormatTypeList.append(self.__formatTypeList[ii])
        except Exception as e:
            logger.exception("Failing with %s " % str(e))

        return curFormatTypeList, curDataTypeList
